from typing import Optional

from flowchem.components.stdlib import Mixer


class YMixer(Mixer):
    """
    A Y mixer.

    This is an alias of `Component`.

    Arguments:
    - `name`: The name of the mixer.

    Attributes:
    - See arguments.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)



from typing import Optional

from loguru import logger
from flowchem.components.stdlib import Valve


class DummyValve(Valve):
    """
    A fake valve, used internally for testing.

    ::: danger
    This component can be used in a real protocol, although it doesn't actually exist.
    :::

    Arguments:
    - `mapping`: The mapping from components to their integer port numbers.
    - `name`: The name of the valve.

    Attributes:
    - `mapping`: The mapping from components to their integer port numbers.
    - `name`: The name of the valve.
    - `setting`: The position of the valve as an int (mapped via `mapping`).
    """

    def __init__(self, name: Optional[str] = None, mapping: dict = {}):
        super().__init__(name=name, mapping=mapping)

    async def _update(self) -> None:
        logger.trace(f"Switching {self.name} to position {self.setting}")



from .component import Component
from .interface import Interface
from .mapped_component import MappedComponentMixin
from .active_component import ActiveComponent
from .dummy import Dummy
from .sensor import Sensor
from .tube import Tube
from .broken_dummy_component import BrokenDummyComponent
from .broken_dummy_sensor import BrokenDummySensor
from .mixer import Mixer
from .cross_mixer import CrossMixer
from .t_mixer import TMixer
from .y_mixer import YMixer
from .pump import Pump
from .dummy_pump import DummyPump
from .dummy_sensor import DummySensor
from .tempcontrol import TempControl
from .valve import Valve
from .dummy_valve import DummyValve
from .vessel import Vessel
from .vessel_chemicals import VesselChemicals



"""
This module is used to control Hamilton ML600 syringe pump via the protocol1/RNO+.
"""

from __future__ import annotations

from loguru import logger
import string
import time
import warnings
from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, Set, TYPE_CHECKING
if TYPE_CHECKING:
    import pint

import aioserial

from flowchem.exceptions import InvalidConfiguration, DeviceError
from flowchem.units import flowchem_ureg


@dataclass
class Protocol1CommandTemplate:
    """Class representing a pump command and its expected reply, but without target pump number"""

    command: str
    optional_parameter: str = ""
    execute_command: bool = True

    def to_pump(
        self, address: int, command_value: str = "", argument_value: str = ""
    ) -> Protocol1Command:
        """Returns a Protocol11Command by adding to the template pump address and command arguments"""
        return Protocol1Command(
            target_pump_num=address,
            command=self.command,
            optional_parameter=self.optional_parameter,
            command_value=command_value,
            argument_value=argument_value,
            execute_command=self.execute_command,
        )


@dataclass
class Protocol1Command(Protocol1CommandTemplate):
    """Class representing a pump command and its expected reply"""

    PUMP_ADDRESS = {
        pump_num: address
        for (pump_num, address) in enumerate(string.ascii_lowercase[:16], start=1)
    }
    # i.e. PUMP_ADDRESS = {1: 'a', 2: 'b', 3: 'c', 4: 'd', ..., 16: 'p'}
    # Note ':' is used for broadcast within the daisy chain.

    target_pump_num: int = 1
    command_value: Optional[str] = None
    argument_value: Optional[str] = None

    def compile(self) -> bytes:
        """Create actual command byte by prepending pump address to command and appending executing command."""
        assert self.target_pump_num in range(1, 17)
        if not self.command_value:
            self.command_value = ""

        compiled_command = (
            f"{self.PUMP_ADDRESS[self.target_pump_num]}"
            f"{self.command}{self.command_value}"
        )

        if self.argument_value:
            compiled_command += f"{self.optional_parameter}{self.argument_value}"
        # Add execution flag at the end
        if self.execute_command is True:
            compiled_command += "R"

        return (compiled_command + "\r").encode("ascii")


class HamiltonPumpIO:
    """Setup with serial parameters, low level IO"""

    ACKNOWLEDGE = chr(6)
    NEGATIVE_ACKNOWLEDGE = chr(21)
    DEFAULT_CONFIG = {
        "timeout": 0.1,
        "baudrate": 9600,
        "parity": aioserial.PARITY_EVEN,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.SEVENBITS,
    }

    def __init__(self, aio_port: aioserial.Serial):
        """
        Initialize communication on the serial port where the pumps are located and initialize them
        Args:
            aio_port: aioserial.Serial() object
        """
        self._serial = aio_port

        # These will be set by `HamiltonPumpIO.initialize()`
        self._initialized = False
        self.num_pump_connected: Optional[int] = None

    @classmethod
    def from_config(cls, config):
        """Create HamiltonPumpIO from config."""
        # Merge default settings, including serial, with provided ones.
        configuration = dict(HamiltonPumpIO.DEFAULT_CONFIG, **config)

        try:
            serial_object = aioserial.AioSerial(**configuration)
        except aioserial.SerialException as e:
            raise InvalidConfiguration(
                f"Cannot connect to the pump on the port <{configuration.get('port')}>"
            ) from e

        return cls(serial_object)

    async def initialize(self, hw_initialization: bool = True):
        """
        Ensure connection with pump + initialize

        Args:
            hw_initialization: Whether each pumps has to be initialized. Note that this might be undesired!
        """
        # This has to be run after each power cycle to assign addresses to pumps
        self.num_pump_connected = await self._assign_pump_address()
        if hw_initialization:
            await self._hw_init()
        self._initialized = True

    async def _assign_pump_address(self) -> int:
        """
        To be run on init, auto assign addresses to pumps based on their position on the daisy chain!
        A custom command syntax with no addresses is used here so read and write has been rewritten
        """
        try:
            await self._write_async("1a\r".encode("ascii"))
        except aioserial.SerialException as e:
            raise InvalidConfiguration from e

        reply = await self._read_reply_async()
        if reply and reply[:1] == "1":
            # reply[1:2] should be the address of the last pump. However, this does not work reliably.
            # So here we enumerate the pumps explicitly instead
            last_pump = 0
            for pump_num, address in Protocol1Command.PUMP_ADDRESS.items():
                await self._write_async(f"{address}UR\r".encode("ascii"))
                if "NV01" in await self._read_reply_async():
                    last_pump = pump_num
                else:
                    break
            logger.debug(f"Found {last_pump} pumps on {self._serial.port}!")
            return int(last_pump)
        else:
            raise InvalidConfiguration(f"No pump found on {self._serial.port}")

    async def _hw_init(self):
        """Send to all pumps the HW initialization command (i.e. homing)"""
        await self._write_async(b":XR\r")  # Broadcast: initialize + execute
        # Note: no need to consume reply here because there is none (since we are using broadcast)

    async def _write_async(self, command: bytes):
        """Writes a command to the pump"""
        if not self._initialized:
            raise DeviceError(
                "Pump not initialized!\n"
                "Have you called `initialize()` after object creation?"
            )
        await self._serial.write_async(command)
        logger.debug(f"Command {repr(command)} sent!")

    async def _read_reply_async(self) -> str:
        """Reads the pump reply from serial communication"""
        reply_string = await self._serial.readline_async()
        logger.debug(f"Reply received: {reply_string}")
        return reply_string.decode("ascii")

    def parse_response(self, response: str) -> str:
        """Split a received line in its components: success, reply"""
        status = response[:1]
        assert status in (
            HamiltonPumpIO.ACKNOWLEDGE,
            HamiltonPumpIO.NEGATIVE_ACKNOWLEDGE,
        ), "Invalid status reply!"

        if status == HamiltonPumpIO.ACKNOWLEDGE:
            logger.debug("Positive acknowledge received")
        else:
            logger.warning("Negative acknowledge received")
            warnings.warn(
                "Negative acknowledge reply received from pump: check command validity!"
            )

        return response[1:].rstrip()

    def reset_buffer(self):
        """Reset input buffer before reading from serial. In theory not necessary if all replies are consumed..."""
        self._serial.reset_input_buffer()

    async def write_and_read_reply_async(self, command: Protocol1Command) -> str:
        """Main HamiltonPumpIO method.
        Sends a command to the pump, read the replies and returns it, optionally parsed"""
        self.reset_buffer()
        await self._write_async(command.compile())
        response = await self._read_reply_async()

        if not response:
            raise InvalidConfiguration(
                f"No response received from pump, check pump address! "
                f"(Currently set to {command.target_pump_num})"
            )

        return self.parse_response(response)

    @property
    def name(self) -> str:
        """This is used to provide a nice-looking default name to pumps based on their serial connection."""
        try:
            return self._serial.name
        except AttributeError:
            return ""


class ML600:
    """ " ML600 implementation according to docs. Tested on 61501-01 (single syringe).

    From docs:
    To determine the volume dispensed per step the total syringe volume is divided by
    48,000 steps. All Hamilton instrument syringes are designed with a 60 mm stroke
    length and the Microlab 600 is designed to move 60 mm in 48,000 steps. For
    example to dispense 9 mL from a 10 mL syringe you would determine the number of
    steps by multiplying 48000 steps (9 mL/10 mL) to get 43,200 steps.
    """

    # This class variable is used for daisy chains (i.e. multiple pumps on the same serial connection). Details below.
    _io_instances: Set[HamiltonPumpIO] = set()
    # The mutable object (a set) as class variable creates a shared state across all the instances.
    # When several pumps are daisy chained on the same serial port, they need to all access the same Serial object,
    # because access to the serial port is exclusive by definition (also locking there ensure thread safe operations).
    # FYI it is a borg idiom https://www.oreilly.com/library/view/python-cookbook/0596001673/ch05s23.html

    class ValvePositionName(IntEnum):
        """Maps valve position to the corresponding number"""

        POSITION_1 = 1
        # POSITION_2 = 2
        POSITION_3 = 3
        INPUT = 9  # 9 is default inlet, i.e. 1
        OUTPUT = 10  # 10 is default outlet, i.e. 3
        WASH = 11  # 11 is default wash, i.e. undefined

    # Only Hamilton syringes are compatible w/ the ML600, and they come on a limited set of sizes. (Values in ml)
    VALID_SYRINGE_VOLUME = {
        0.01,
        0.025,
        0.05,
        0.1,
        0.25,
        0.5,
        1.0,
        2.5,
        5.0,
        10.0,
        25.0,
        50.0,
    }

    def __init__(
        self,
        pump_io: HamiltonPumpIO,
        syringe_volume: str,
        address: int = 1,
        name: str = None,
    ):
        """
        Default constructor, needs an HamiltonPumpIO object. See from_config() class method for config-based init.

        Args:
            pump_io: An HamiltonPumpIO w/ serial connection to the daisy chain w/ target pump.
            syringe_volume: Volume of the syringe used, either a Quantity or number in ml.
            address: number of pump in array, 1 for first one, auto-assigned on init based on position.
            name: 'cause naming stuff is important.
        """
        # HamiltonPumpIO
        self.pump_io = pump_io
        ML600._io_instances.add(self.pump_io)  # See above for details.

        # Pump address is the pump sequence number if in chain. Count starts at 1, default.
        self.address = int(address)

        # The pump name is used for logs and error messages.
        self.name = f"Pump {self.pump_io.name}:{address}" if name is None else name

        # Syringe pumps only perform linear movement, and the volume displaced is function of the syringe loaded.
        self.syringe_volume = flowchem_ureg(syringe_volume)

        if self.syringe_volume.m_as("ml") not in ML600.VALID_SYRINGE_VOLUME:
            raise InvalidConfiguration(
                f"The specified syringe volume ({syringe_volume}) does not seem to be valid!\n"
                f"The volume in ml has to be one of {ML600.VALID_SYRINGE_VOLUME}"
            )

        self._steps_per_ml = flowchem_ureg.Quantity(
            f"{48000 / self.syringe_volume} step/ml"
        )
        self._offset_steps = 100  # Steps added to each absolute move command, to decrease wear and tear at volume = 0
        self._max_vol = (
            (48000 - self._offset_steps) * flowchem_ureg.step / self._steps_per_ml
        )

    @classmethod
    def from_config(cls, config):
        """This class method is used to create instances via config file by the server for HTTP interface."""
        # Many pump can be present on the same serial port with different addresses.
        # This shared list of HamiltonPumpIO objects allow shared state in a borg-inspired way, avoiding singletons
        # This is only relevant to programmatic instantiation, i.e. when from_config() is called per each pump from a
        # config file, as it is the case in the HTTP server.
        # HamiltonPump_IO() manually instantiated are not accounted for.
        pumpio = None
        for obj in ML600._io_instances:
            # noinspection PyProtectedMember
            if obj._serial.port == config.get("port"):
                pumpio = obj
                break

        # If not existing serial object are available for the port provided, create a new one
        if pumpio is None:
            # Remove ML600-specific keys to only have HamiltonPumpIO's kwargs
            config_for_pumpio = {
                k: v
                for k, v in config.items()
                if k not in ("syringe_volume", "address", "name")
            }
            pumpio = HamiltonPumpIO.from_config(config_for_pumpio)

        return cls(
            pumpio,
            syringe_volume=config.get("syringe_volume"),
            address=config.get("address"),
            name=config.get("name"),
        )

    async def initialize(self, hw_init=False, init_speed: str = "200 sec / stroke"):
        """Must be called after init before anything else."""
        # Test connectivity by querying the pump's firmware version
        fw_cmd = Protocol1CommandTemplate(command="U").to_pump(self.address)
        firmware_version = await self.pump_io.write_and_read_reply_async(fw_cmd)
        logger.info(
            f"Connected to Hamilton ML600 {self.name} - FW version: {firmware_version}!"
        )

        if hw_init:
            await self.initialize_pump(speed=init_speed)

    async def send_command_and_read_reply(
        self,
        command_template: Protocol1CommandTemplate,
        command_value="",
        argument_value="",
    ) -> str:
        """Sends a command based on its template by adding pump address and parameters, returns reply"""
        return await self.pump_io.write_and_read_reply_async(
            command_template.to_pump(self.address, command_value, argument_value)
        )

    def _validate_speed(self, speed_value: Optional[str]) -> str:
        """Given a speed (seconds/stroke) returns a valid value for it, and a warning if out of bounds."""

        # Validated speeds are used as command argument, with empty string being the default for None
        if speed_value is None:
            return ""

        speed = flowchem_ureg(speed_value)

        # Alert if out of bounds but don't raise exceptions, according to general philosophy.
        # Target flow rate too high
        if speed < flowchem_ureg("2 sec/stroke"):
            speed = flowchem_ureg("2 sec/stroke")
            warnings.warn(
                f"Desired speed ({speed}) is unachievable!"
                f"Set to {self._seconds_per_stroke_to_flowrate(speed)}"
                f"Wrong units? A bigger syringe is needed?"
            )

        # Target flow rate too low
        if speed > flowchem_ureg("3692 sec/stroke"):
            speed = flowchem_ureg("3692 sec/stroke")
            warnings.warn(
                f"Desired speed ({speed}) is unachievable!"
                f"Set to {self._seconds_per_stroke_to_flowrate(speed)}"
                f"Wrong units? A smaller syringe is needed?"
            )

        return str(round(speed.m_as("sec / stroke")))

    async def initialize_pump(self, speed: Optional[str] = None):
        """
        Initialize both syringe and valve
        speed: 2-3692 in seconds/stroke
        """
        init_cmd = Protocol1CommandTemplate(command="X", optional_parameter="S")
        return await self.send_command_and_read_reply(
            init_cmd, argument_value=self._validate_speed(speed)
        )

    async def initialize_valve(self):
        """Initialize valve only"""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="LX")
        )

    async def initialize_syringe(self, speed: Optional[str] = None):
        """
        Initialize syringe only
        speed: 2-3692 in seconds/stroke
        """
        init_syringe_cmd = Protocol1CommandTemplate(
            command="X1", optional_parameter="S"
        )
        return await self.send_command_and_read_reply(
            init_syringe_cmd, argument_value=self._validate_speed(speed)
        )

    def flowrate_to_seconds_per_stroke(self, flowrate: str):
        """
        Convert flow rates to steps per seconds

        To determine the volume dispensed per step the total syringe volume is divided by
        48,000 steps. All Hamilton instrument syringes are designed with a 60 mm stroke
        length and the Microlab 600 is designed to move 60 mm in 48,000 steps. For
        example to dispense 9 mL from a 10 mL syringe you would determine the number of
        steps by multiplying 48000 steps (9 mL/10 mL) to get 43,200 steps.
        """
        flowrate = flowchem_ureg(flowrate)
        flowrate_in_steps_sec = flowrate * self._steps_per_ml
        seconds_per_stroke = (1 / flowrate_in_steps_sec).to("second/stroke")

        return self._validate_speed(str(seconds_per_stroke))

    def _seconds_per_stroke_to_flowrate(self, second_per_stroke: pint.Quantity) -> float:
        """ The inverse of flowrate_to_seconds_per_stroke(). Only internal use. """
        flowrate = 1 / (second_per_stroke * self._steps_per_ml)
        return flowrate.to("ml/min")

    def _volume_to_step_position(self, volume_w_units: str) -> int:
        """ Converts a volume to a step position. """
        # noinspection PyArgumentEqualDefault
        volume = flowchem_ureg(volume_w_units)
        steps = volume * self._steps_per_ml
        return round(steps.m_as("steps")) + self._offset_steps

    async def _to_step_position(self, position: int, speed: str = ""):
        """ Absolute move to step position. """
        abs_move_cmd = Protocol1CommandTemplate(command="M", optional_parameter="S")
        return await self.send_command_and_read_reply(
            abs_move_cmd, str(position), self._validate_speed(speed)
        )

    async def get_current_volume(self) -> str:
        """Return current syringe position in ml."""
        syringe_pos = await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="YQP")
        )
        current_steps = (int(syringe_pos) - self._offset_steps) * flowchem_ureg.step
        return str(current_steps / self._steps_per_ml)

    async def to_volume(self, target_volume: str, speed: str = ""):
        """ Absolute move to volume provided. """
        await self._to_step_position(self._volume_to_step_position(target_volume), speed)
        logger.debug(
            f"Pump {self.name} set to volume {target_volume} at speed {speed}"
        )

    async def pause(self):
        """Pause any running command."""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="K", execute_command=False)
        )

    async def resume(self):
        """Resume any paused command."""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="$", execute_command=False)
        )

    async def stop(self):
        """Stops and abort any running command."""
        await self.pause()
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="V", execute_command=False)
        )

    async def wait_until_idle(self):
        """Returns when no more commands are present in the pump buffer."""
        logger.debug(f"ML600 pump {self.name} wait until idle...")
        while self.is_busy:
            time.sleep(0.1)
        logger.debug(f"...ML600 pump {self.name} idle now!")

    async def version(self) -> str:
        """Returns the current firmware version reported by the pump."""
        return await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="U")
        )

    async def is_idle(self) -> bool:
        """Checks if the pump is idle (actually check if the last command has ended)."""
        return (
            await self.send_command_and_read_reply(
                Protocol1CommandTemplate(command="F")
            )
            == "Y"
        )

    async def is_busy(self) -> bool:
        """Pump is not idle."""
        return not await self.is_idle()

    async def get_valve_position(self) -> ValvePositionName:
        """Represent the position of the valve: getter returns Enum, setter needs Enum."""
        valve_pos = await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="LQP")
        )
        return ML600.ValvePositionName(int(valve_pos))

    async def set_valve_position(
        self, target_position: ValvePositionName, wait_for_movement_end: bool = True
    ):
        """Set valve position. wait_for_movement_end is defaulted to True as it is a common mistake not to wait..."""
        valve_by_name_cw = Protocol1CommandTemplate(command="LP0")
        await self.send_command_and_read_reply(
            valve_by_name_cw, command_value=str(int(target_position))
        )
        logger.debug(f"{self.name} valve position set to {target_position.name}")
        if wait_for_movement_end:
            await self.wait_until_idle()

    async def get_return_steps(self) -> int:
        """Return steps getter. Applied to the end of a downward syringe movement to removes mechanical slack."""
        steps = await self.send_command_and_read_reply(
            Protocol1CommandTemplate(command="YQN")
        )
        return int(steps)

    async def set_return_steps(self, target_steps: int):
        """Return steps setter. Applied to the end of a downward syringe movement to removes mechanical slack."""
        set_return_steps_cmd = Protocol1CommandTemplate(command="YSN")
        await self.send_command_and_read_reply(
            set_return_steps_cmd, command_value=str(int(target_steps))
        )

    async def pickup(
        self,
        volume: str,
        from_valve: ValvePositionName,
        flowrate: str = "1 ml/min",
        wait: bool = False,
    ):
        """Get volume from valve specified at given flowrate."""
        cur_vol = flowchem_ureg(await self.get_current_volume())
        if (cur_vol + volume) > self._max_vol:
            warnings.warn(
                f"Cannot withdraw {volume} given the current syringe position {cur_vol} and a "
                f"syringe volume of {self.syringe_volume}"
            )
            return

        # Valve to position specified
        await self.set_valve_position(from_valve, wait_for_movement_end=True)
        # Move up to target volume
        await self.to_volume(
            str(cur_vol + volume),
            speed=self.flowrate_to_seconds_per_stroke(flowrate),
        )

        if wait:
            await self.wait_until_idle()

    async def deliver(
        self,
        volume: str,
        to_valve: ValvePositionName,
        flowrate: str,
        wait: bool = False,
    ):
        """Delivers volume to valve specified at given flow rate."""
        cur_vol = flowchem_ureg(await self.get_current_volume())
        if volume > cur_vol:
            warnings.warn(
                f"Cannot deliver {volume} given the current syringe position {cur_vol}!"
            )
            return

        # Valve to position specified
        await self.set_valve_position(to_valve, wait_for_movement_end=True)
        # Move up to target volume
        await self.to_volume(
            str(cur_vol - volume),
            speed=self.flowrate_to_seconds_per_stroke(flowrate),
        )

        if wait:
            await self.wait_until_idle()

    async def transfer(self, volume: str, from_valve: ValvePositionName, to_valve: ValvePositionName,
                       flowrate_in: str = "1 ml/min", flowrate_out: str = "1 ml/min", wait: bool = False):
        """Move liquid from place to place."""
        await self.pickup(volume, from_valve, flowrate_in, wait=True)
        await self.deliver(volume, to_valve, flowrate_out, wait=wait)

    def get_router(self):
        """Creates an APIRouter for this object."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/firmware-version", self.version, methods=["GET"])
        router.add_api_route("/initialize/pump", self.initialize_pump, methods=["PUT"])
        router.add_api_route(
            "/initialize/valve", self.initialize_valve, methods=["PUT"]
        )
        router.add_api_route(
            "/initialize/syringe", self.initialize_syringe, methods=["PUT"]
        )
        router.add_api_route("/pause", self.pause, methods=["PUT"])
        router.add_api_route("/resume", self.resume, methods=["PUT"])
        router.add_api_route("/resume", self.resume, methods=["PUT"])
        router.add_api_route("/stop", self.stop, methods=["PUT"])
        router.add_api_route("/version", self.stop, methods=["PUT"])
        router.add_api_route("/is-idle", self.is_idle, methods=["GET"])
        router.add_api_route("/is-busy", self.is_busy, methods=["GET"])
        router.add_api_route(
            "/valve/position", self.get_valve_position, methods=["GET"]
        )
        router.add_api_route(
            "/valve/position", self.set_valve_position, methods=["PUT"]
        )
        router.add_api_route(
            "/syringe/volume", self.get_current_volume, methods=["GET"]
        )
        router.add_api_route("/syringe/volume", self.to_volume, methods=["PUT"])
        router.add_api_route(
            "/syringe/return-steps", self.get_return_steps, methods=["GET"]
        )
        router.add_api_route(
            "/syringe/return-steps", self.set_return_steps, methods=["PUT"]
        )
        router.add_api_route("/pickup", self.pickup, methods=["PUT"])
        router.add_api_route("/deliver", self.deliver, methods=["PUT"])
        # router.add_api_route("/transfer", self.transfer, methods=["PUT"])  # Might go in timeout

        return router


# class TwoPumpAssembly(Thread):
#     """
#     Thread to control two pumps and have them generating a continuous flow.
#     Note that the pumps should not be accessed directly when used in a TwoPumpAssembly!
#
#     Notes: this needs to start a thread owned by the instance to control the pumps.
#     The async version of this being possibly simpler w/ tasks and callback :)
#     """
#
#     def __init__(
#         self, pump1: ML600, pump2: ML600, target_flowrate: str, init_seconds: int = 10
#     ):
#         super(TwoPumpAssembly, self).__init__()
#         self._p1 = pump1
#         self._p2 = pump2
#         self.daemon = True
#         self.cancelled = threading.Event()
#         self._flowrate = ensure_quantity(target_flowrate, "ml/min")
#         logger = logging.getLogger(__name__).getChild("TwoPumpAssembly")
#         # How many seconds per stroke for first filling? application dependent, as fast as possible, but not too much.
#         self.init_secs = init_seconds
#
#         # While in principle possible, using syringes of different volumes is discouraged, hence...
#         assert (
#             pump1.syringe_volume == pump2.syringe_volume
#         ), "Syringes w/ equal volume are needed for continuous flow!"
#
#     async def initialize(self):
#         """ Initialize multi-pump """
#         await self._p1.initialize()
#         await self._p2.initialize()
#
#     @property
#     def flowrate(self):
#         """ Returns/sets flowrate. """
#         return self._flowrate
#
#     @flowrate.setter
#     def flowrate(self, target_flowrate):
#         if target_flowrate == 0:
#             warnings.warn(
#                 "Cannot set flowrate to 0! Pump stopped instead, restart previous flowrate with resume!"
#             )
#             self.cancel()
#         else:
#             self._flowrate = target_flowrate
#
#         # This will stop current movement, make wait_for_both_pumps() return and move on w/ updated speed
#         self._p1.stop()
#         self._p2.stop()
#
#     async def wait_for_both_pumps(self):
#         """ Custom waiting method to wait a shorter time than normal (for better sync) """
#         while await self._p1.is_busy() or await self._p2.is_busy():
#             await asyncio.sleep(0.01)  # 10ms sounds reasonable to me
#         logger.debug("Both pumps are ready!")
#
#     def _speed(self):
#         speed = self._p1.flowrate_to_seconds_per_stroke(self._flowrate)
#         logger.debug(f"Speed calculated as {speed}")
#         return speed
#
#     async def execute_stroke(
#         self, pump_full: ML600, pump_empty: ML600, speed_s_per_stroke: int
#     ):
#         """ Perform a cycle (1 syringe stroke) in the continuous-operation mode. See also run(). """
#         # Logic is a bit complex here to ensure pause-less pumping
#         # This needs the pump that withdraws to move faster than the pumping one. no way around.
#
#         # First start pumping with the full syringe already prepared
#         pump_full.to_volume(0, speed=speed_s_per_stroke)
#         logger.debug("Pumping...")
#         # Then start refilling the empty one
#         pump_empty.set_valve_position(pump_empty.ValvePositionName.INPUT)
#         # And do that fast so that we finish refill before the pumping is over
#         pump_empty.to_volume(pump_empty.syringe_volume, speed=speed_s_per_stroke - 5)
#         pump_empty.wait_until_idle()
#         # This allows us to set the right pump position on the pump that was empty (not full and ready for next cycle)
#         pump_empty.set_valve_position(pump_empty.ValvePositionName.OUTPUT)
#         pump_full.wait_until_idle()
#
#     def run(self):
#         """Overloaded Thread.run, runs the update
#         method once per every 10 milliseconds."""
#         # First initialize with init_secs speed...
#         self._p1.to_volume(self._p1.syringe_volume, speed=self.init_secs)
#         self._p1.wait_until_idle()
#         self._p1.valve_position = self._p1.ValvePositionName.OUTPUT
#         logger.info("Pumps initialized for continuous pumping!")
#
#         while True:
#             while not self.cancelled.is_set():
#                 self.execute_stroke(
#                     self._p1, self._p2, speed_s_per_stroke=self._speed()
#                 )
#                 self.execute_stroke(
#                     self._p2, self._p1, speed_s_per_stroke=self._speed()
#                 )
#
#     def cancel(self):
#         """ Cancel continuous-pumping assembly """
#         self.cancelled.set()
#         self._p1.stop()
#         self._p2.stop()
#
#     def resume(self):
#         """ Resume continuous-pumping assembly """
#         self.cancelled.clear()
#
#     def stop_and_return_solution_to_container(self):
#         """ Let´s not waste our precious stock solutions ;) """
#         self.cancel()
#         logger.info(
#             "Returning the solution currently loaded in the syringes back to the inlet.\n"
#             "Make sure the container is not removed yet!"
#         )
#         # Valve to input
#         self._p1.valve_position = self._p1.ValvePositionName.INPUT
#         self._p2.valve_position = self._p2.ValvePositionName.INPUT
#         self.wait_for_both_pumps()
#         # Volume to 0 with the init speed (supposedly safe for this application)
#         self._p1.to_volume(0, speed=self.init_secs)
#         self._p2.to_volume(0, speed=self.init_secs)
#         self.wait_for_both_pumps()
#         logger.info("Pump flushing completed!")


if __name__ == "__main__":
    import asyncio

    conf = {
        "port": "COM12",
        "address": 1,
        "name": "test1",
        "syringe_volume": 5,
    }
    pump1 = ML600.from_config(conf)
    asyncio.run(pump1.initialize_pump())



"""
This module is used to discover the serial address of any ML600 connected to the PC.
"""
from loguru import logger

import serial.tools.list_ports

from flowchem.components.devices.Harvard_Apparatus.HA_elite11 import (
    HarvardApparatusPumpIO,
)
from flowchem.exceptions import InvalidConfiguration


# noinspection PyProtectedMember
def elite11_finder():
    """Try to initialize an Elite11 on every available COM port."""
    port_available = [comport.device for comport in serial.tools.list_ports.comports()]

    # Ports connected to an elite11-looking device
    valid_ports = set()

    for serial_port in port_available:
        try:
            print(f"Looking for pump on {serial_port}...")
            link = HarvardApparatusPumpIO(port=serial_port)
            link._serial.write("\r\n".encode("ascii"))
            if link._serial.readline() == b"\n":
                valid_ports.add(serial_port)
                logger.info(f"Pump found on <{serial_port}>")
                pump = link._serial.readline().decode("ascii")
                logger.info(f"Pump address is {pump[0:2]}!")
                print(f"Found a pump with address {pump[0:2]} on {serial_port}!")
            else:
                logger.debug(f"No pump found on {serial_port}")
        except InvalidConfiguration:
            pass

    return valid_ports


if __name__ == "__main__":
    elite11_pumps = elite11_finder()
    if len(elite11_pumps) > 0:
        print(f"The following serial port are connected to Elite11: {elite11_pumps}")
    else:
        print("No Elite11 pump found")



"""" Run with uvicorn main:app """
from pathlib import Path
from loguru import logger
from typing import Dict, Tuple

import jsonschema
import yaml
from fastapi import FastAPI
from fastapi.responses import HTMLResponse

import flowchem
from flowchem.core.graph.DeviceGraph import (
    DEVICE_MODULES,
    get_device_class_mapper,
    load_schema,
)
from flowchem.core.graph.DeviceNode import DeviceNode
from flowchem.exceptions import InvalidConfiguration
from mdns_server import Server_mDNS


def create_server_from_config(
    config: Dict = None, config_file: Path = None
) -> Tuple[FastAPI, Server_mDNS]:
    """
    Based on the yaml device graph provided, creates device objects and connect to them + .

    config: Path to the yaml file with the device config or dict.
    """

    assert (
        config is not None
        and config_file is None
        or config is None
        and config_file is not None
    )

    if config_file is not None:
        with config_file.open() as stream:
            config = yaml.safe_load(stream)

    assert isinstance(config, dict)  # This is here just to make mypy happy.

    # Validate config
    schema = load_schema()
    jsonschema.validate(config, schema=schema)

    # FastAPI server
    app = FastAPI(title="flowchem", version=flowchem.__version__)

    # Zeroconf server
    zeroconf = Server_mDNS()

    # Device mapper
    device_mapper = get_device_class_mapper(DEVICE_MODULES)
    logger.debug(
        f"The following device classes have been found: {device_mapper.keys()}"
    )

    # Parse list of devices and generate endpoints
    for device_name, node_config in config["devices"].items():
        # Schema validation ensures only 1 hit here
        try:
            device_class = [
                name for name in device_mapper.keys() if name in node_config
            ].pop()
        except IndexError as e:
            raise InvalidConfiguration(f"No class available for device '{device_name}'") from e

        # Object type
        obj_type = device_mapper[device_class]
        device_config = node_config[device_class]

        node = DeviceNode(device_name, device_config, obj_type)
        logger.debug(f"Created device <{device_name}> with config: {device_config}")

        # Add to App
        app.include_router(node.router, prefix=node.router.prefix, tags=node.router.tags)
        logger.debug(f"Router for <{device_name}> added to app!")

        # Add to mDNS server
        zeroconf.include_device(node.safe_title, node.router.prefix)
        logger.debug(f"Router for <{device_name}> added to app!")

    return app, zeroconf


if __name__ == "__main__":
    app, zeroconf = create_server_from_config(
        config_file=Path("../graph/flowchem-graph.yml")
    )

    @app.get("/", response_class=HTMLResponse, include_in_schema=False)
    def root():
        """Server root"""
        # FIXME add landing page
        return "<h1>Flowchem Device Server!</h1>" \
               "<a href='./docs/'>API Reference</a>"

    import uvicorn

    uvi = uvicorn.run(app, host="127.0.0.1")



import random
from typing import Optional
from flowchem.components.stdlib import Sensor


class DummySensor(Sensor):
    """A dummy sensor returning the number of times it has been read.

    ::: danger
    Don't use this in a real apparatus! It doesn't return real data.
    :::

    Attributes:
    - `name`: The component's name.
    - `rate`: Data collection rate in Hz as a `pint.Quantity`. A rate of 0 Hz corresponds to the sensor being off.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self._unit = "Dimensionless"
        self.counter = 0.0

    async def _read(self) -> float:
        """Collect the data."""
        self.counter += (random.random() * 2) - 1
        return self.counter



""" Connection reader + XML parser for Spinsolve replies """

import time
import warnings
import queue
from typing import List

from lxml import etree

from flowchem.components.devices.Magritek.utils import get_my_docs_path


class Reader:
    """
    This class is responsible for collecting and parsing replies from the spectrometer.
    It does not contain functionality to handle I/O.
    """

    def __init__(self, reply_queue: queue.Queue, xml_schema=None):
        self._queue = reply_queue

        if xml_schema is None:
            # This is the default location upon Spinsolve installation. However, remote control can be from remote ;)
            my_docs = get_my_docs_path()
            try:
                self.schema = etree.XMLSchema(
                    file=str(my_docs / "Magritek" / "Spinsolve" / "RemoteControl.xsd")
                )
            except etree.XMLSchemaParseError:  # i.e. not found
                self.schema = None
        else:
            self.schema = xml_schema

        self.parser = etree.XMLParser()
        self._replies: List[etree.Element] = []
        self._rcv_buffer = b""

    def wait_for_reply(self, reply_type="", remove=True, timeout=1):
        """
        Awaits for a reply of type reply_type or up to timeout
        """
        reply = self.get_last_reply(reply_type, remove)

        # If already available just return
        if reply is not None:
            return reply

        # This is ugly, but usually unnecessary as replies are received immediately.
        # Only relevant if controlling remote devices over connections with significant latency
        start_time = time.time()
        while reply is None and time.time() < (start_time + timeout):
            reply = self.get_last_reply(reply_type)
            time.sleep(0.5)

        if reply is None:
            raise RuntimeError("No reply received from device!")

        return reply

    def get_last_reply(self, reply_type="", remove=True):
        """
        Returns the last reply of given type
        """
        self.fetch_replies()

        for reply in reversed(self._replies):
            # First tag in message is response type
            if reply[0].tag.endswith(reply_type):
                if remove:
                    self._replies.remove(
                        reply
                    )  # yes, I am also surprised that this is possible ;)
                return reply

    def clear_replies(self, reply_type=""):
        """Remove old replies."""
        # Shortcut if none provided
        if not reply_type:
            self._replies.clear()

        # Otherwise check type
        for reply in self._replies:
            if reply[0].tag.endswith(reply_type):
                self._replies.remove(reply)

    def fetch_replies(self):
        """
        Fetch the unprocessed chunks from the queue and adds them to the receive buffer
        """
        while not self._queue.empty():
            self._rcv_buffer += self._queue.get()
            self._queue.task_done()

        self.parse_buffer()

    def parse_buffer(self):
        """
        Split then buffer into individual XML trees and parse them.
        """

        # Split buffer at each declaration to separate different trees
        declaration = b'<?xml version="1.0" encoding="utf-8"?>'
        trees = self._rcv_buffer.split(declaration)

        # Parse trees. (Buffer always starts with declaration, so split gives a tree[0] = b"" element that we skip
        for tree in trees[1:]:
            self.parse_tree(tree)

        # Clear buffer
        self._rcv_buffer = b""

    def parse_tree(self, tree_string):
        """
        Parse an XML reply tree, add it to the replies and validate it (if the schema is available).
        """

        # Parse
        try:
            root = etree.fromstring(tree_string, self.parser)
        except etree.XMLSyntaxError:
            warnings.warn(f"Cannot parse response XML {tree_string}")
            return None

        # Add tree to replies
        self._replies.append(root)

        # Validate
        if self.schema:
            try:
                self.schema.validate(root)
            except etree.XMLSyntaxError as e:
                warnings.warn(f"Invalid XML received! [Validation error: {e}]")



""" A vessel with info on the chemical input contained. """
from ord_schema.proto.reaction_pb2 import ReactionInput
from typing import Optional
from flowchem.components.stdlib import Vessel


class VesselChemicals(Vessel):
    """
    A Vessel with additional properties for chemical identity
    """

    metadata = {
        "author": [
            {
                "first_name": "Dario",
                "last_name": "Cambie",
                "email": "dario.cambie@mpikg.mpg.de",
                "institution": "Max Planck Institute of Colloids and Interfaces",
                "github_username": "dcambie",
            }
        ],
        "stability": "beta",
        "supported": True,
    }

    def __init__(self, reaction_input: ReactionInput, description: Optional[str] = None, name: Optional[str] = None):
        super().__init__(name=name, description=description)
        self.chemical = reaction_input

    def _validate(self, dry_run):
        super(VesselChemicals, self)._validate(dry_run)

        assert isinstance(self.chemical, ReactionInput), "VesselChemicals have a ReactionInput"



from typing import Callable, Union, Tuple
import asyncio
import ctypes.wintypes
import warnings
from pathlib import Path


def get_my_docs_path():
    """
    Spinsolve control software is only available on Windows, so lack of cross-platform support is unavoidable.
    XSD and XML schema are installed in my documents, whose location, if custom, can be obtained as follows.
    """
    # From https://stackoverflow.com/questions/6227590

    csidl_personal = 5  # My Documents
    shgfp_type_current = 0  # Get current, not default value
    buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
    ctypes.windll.shell32.SHGetFolderPathW(
        None, csidl_personal, None, shgfp_type_current, buf
    )
    return Path(buf.value)


def create_folder_mapper(
    remote_root: Path, local_root: Path
) -> Callable[[Union[Path, str]], Path]:
    """
    Returns a function that converts path relative to remote_root to their corresponding on local_root
    Used when using spinsolve on a remote PC to share the result data via a remotely mounted network drive
    """

    def folder_mapper(path_to_be_translated: Union[Path, str]):
        """
        Given a remote path converts it to the corresponding local location, or None + warning if not possible.
        """
        # Ensures it is a Path object
        if isinstance(path_to_be_translated, str):
            path_to_be_translated = Path(path_to_be_translated)

        nonlocal remote_root, local_root
        # If relative translate if not error
        # NOTE: Path.is_relative_to() is available from Py 3.9 only. NBD as this is not often used.
        if path_to_be_translated.is_relative_to(remote_root):
            suffix = path_to_be_translated.relative_to(remote_root)
            return local_root / suffix
        else:
            warnings.warn(
                f"Cannot translate remote path {path_to_be_translated} to a local path!"
                f"{path_to_be_translated} is not relative to {remote_root}"
            )

    return folder_mapper


async def get_streams_for_connection(
    host, port
) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    """
    Given a target (host, port) returns the corresponding asyncio streams (I/O).
    """
    try:
        read, write = await asyncio.open_connection(host, port)
    except Exception as e:
        raise ConnectionError(f"Error connecting to {host}:{port} -- {e}") from e
    return read, write


def run_loop(loop):
    """

    Args:
        loop:
    """
    asyncio.set_event_loop(loop)
    loop.run_forever()



from __future__ import annotations
import asyncio
import time
from typing import Optional, AsyncGenerator, TYPE_CHECKING
from warnings import warn

from loguru import logger

from flowchem.units import flowchem_ureg
from flowchem.components.stdlib import ActiveComponent

if TYPE_CHECKING:
    from flowchem import Experiment


class Sensor(ActiveComponent):
    """
    A generic sensor.

    Attributes:
    - `name`: The name of the Sensor.
    - `rate`: Data collection rate in Hz as a `pint.Quantity`. A rate of 0 Hz corresponds to the sensor being off.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self.rate = flowchem_ureg.parse_expression("0 Hz")
        self._visualization_shape = "ellipse"
        self._unit: str = ""
        self._base_state = {"rate": "0 Hz"}

    async def _read(self):
        """
        Collects the data.
        In the generic `Sensor` implementation, this raises a `NotImplementedError`.
        Subclasses of `Sensor` should implement their own version of this method.
        """
        raise NotImplementedError

    async def _monitor(
        self, experiment: "Experiment", dry_run: bool = False
    ) -> AsyncGenerator:
        """
        If data collection is off and needs to be turned on, turn it on.
        If data collection is on and needs to be turned off, turn off and return data.
        """
        while not experiment._end_loop:  # type: ignore
            # if the sensor is off, hand control back over
            if not self.rate:
                await asyncio.sleep(0)
                continue

            if not dry_run:
                yield {"data": await self._read(), "timestamp": time.time()}
            else:
                yield {"data": "simulated read", "timestamp": time.time()}

            # then wait for the sensor's next read
            if self.rate:
                await asyncio.sleep(1 / self.rate.to_base_units().magnitude)

        logger.debug(f"Monitor loop for {self} has completed.")

    async def _validate_read(self):
        async with self:
            logger.trace("Context entered")
            res = await self._read()
            if not res:
                warn(
                    "Sensor reads should probably return data. "
                    f"Currently, {self}._read() does not return anything."
                )

    def _validate(self, dry_run: bool) -> None:
        logger.debug(f"Performing sensor specific checks for {self}...")
        if not dry_run:
            logger.trace("Executing Sensor-specific checks...")
            logger.trace("Entering context...")
            asyncio.run(self._validate_read())
        logger.trace("Performing general component checks...")
        super()._validate(dry_run=dry_run)

    async def _update(self) -> None:
        # sensors don't have an update method; they implement read
        pass



from typing import Set, Optional


class Component(object):
    """
    One of the individual, irreducible parts of a flow chemistry setup.

    All components in an `Apparatus` must be of type `Component`.
    However, it is unlikely that a user will directly instantiate a `Component`.

    Arguments:
    - `name`: The name of the component.

    Attributes:
    - `name`: The name of the component.
    """

    _id_counter = 0
    _used_names: Set[str] = set()

    def __init__(self, name: Optional[str] = None):
        # name the object, either sequentially or with a given name
        if name is None:
            self.name = self.__class__.__name__ + "_" + str(self.__class__._id_counter)
            self.__class__._id_counter += 1
        else:
            self.name = str(name)
        self._visualization_shape = "box"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    def __str__(self):
        return f"{self.__class__.__name__} {self.name}"

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    def _validate(self, dry_run):
        """Components are valid for dry runs, but not for real runs."""
        if not dry_run:
            raise RuntimeError



"""
Module for communication with Knauer pumps and valves.
"""
import asyncio
from loguru import logger

from flowchem.components.devices.Knauer.Knauer_autodiscover import autodiscover_knauer
from flowchem.exceptions import InvalidConfiguration


class KnauerEthernetDevice:
    """
    Common base class for shared logic across Knauer pumps and valves.
    """

    TCP_PORT = 10001
    BUFFER_SIZE = 1024
    _id_counter = 0

    def __init__(self, ip_address, mac_address, name: str = None):
        """
        Knauer Ethernet Device - either pump or valve.

        If a MAC address is given, it is used to autodiscover the IP address.
        Otherwise, the IP address must be given.

        Note that for configuration files, the MAC address is preferred as it is static.

        Args:
            ip_address: IP address of Knauer device
            mac_address: MAC address of Knauer device
            name: name of device (optional)
        """
        super().__init__(name)  # type: ignore

        # MAC address
        if mac_address:
            self.ip_address = self._ip_from_mac(mac_address)
        else:
            self.ip_address = ip_address

        # These will be set in initialize()
        self._reader: asyncio.StreamReader = None  # type: ignore
        self._writer: asyncio.StreamWriter = None  # type: ignore

        # Note: the pump requires "\n\r" as EOL, the valves "\r\n"! So this is set by sublcasses
        self.eol = b""

    def _ip_from_mac(self, mac_address: str) -> str:
        """Gets IP from MAC."""
        # Autodiscover IP from MAC address
        available_devices = autodiscover_knauer()
        # IP if found, None otherwise
        ip_address = available_devices.get(mac_address)
        if ip_address is None:
            raise InvalidConfiguration(
                f"Device with MAC address={mac_address} not found!\n"
                f"[Available: {available_devices}]"
            )
        return ip_address

    async def initialize(self):
        """Initialize connection"""
        try:
            self._reader, self._writer = await asyncio.open_connection(
                host=self.ip_address, port=10001
            )
        except ConnectionError as e:
            raise InvalidConfiguration(
                f"Cannot open connection with Knauer Device IP={self.ip_address}"
            ) from e

    async def _send_and_receive(self, message: str) -> str:
        self._writer.write(message.encode("ascii") + self.eol)
        logger.debug(f"WRITE >>> '{message}' ")
        reply = await self._reader.readuntil(separator=b"\r")
        logger.debug(f"READ <<< '{reply.decode().strip()}' ")
        return reply.decode("ascii").strip()



# This could become a mess...
# what needs to be done is switch the lamps on, which works over serial.
# the rest is just sending commands to the console, possibly also to another machine

# https://www.dataapex.com/documentation/Content/Help/110-technical-specifications/110.020-command-line-parameters/110.020-command-line-parameters.htm?Highlight=command%20line

import tenacity
import subprocess
import socket
from threading import Thread
from typing import Union
from time import sleep
from pathlib import Path

# Todo should have a command constructor dataclass, would be more neat. For now, will do without to get it running asap

# TODO Very weird, when starting from synthesis, fractioning valve is blocked. no idea why, it's ip is not used.

class ClarityInterface:
    def __init__(self, remote: bool = False, host: str = None, port: int = None, path_to_executable: str = None,
                 instrument_number: int = 1):
        # just determine path to executable, and open socket if for remote usage
        self.remote = remote
        self.instrument = instrument_number
        self.path_to_executable = path_to_executable
        if self.remote:
            self.interface = MessageSender(host, port)
            self.command_executor = self.interface.open_socket_and_send
        else:
            self.command_executor = ClarityExecutioner.execute_command  # type:ignore

# TODO would have to have some way to fail
    @classmethod
    def from_config(cls, config_dict: dict):
        try:
            pass
        except:
            pass


    # if remote execute everything on other PC, else on this
    # Todo doesn't make sense here, done other way
    def execute_command(self, command_string):
        if self.remote:
            self.command_executor(command_string)
        else:
            self.command_executor(command_string, self.path_to_executable)

    #bit displaced convenience function to switch on the lamps of hplc detector. Careful, NDA
    # TODO remove if published
    def switch_lamp_on(self, address='192.168.10.111', port=10001):
        """
        Has to be performed BEFORE starting clarity, otherwise sockets get blocked
        Args:
            address:
            port:

        Returns:

        """
        # send the  respective two commands and check return. Send to socket
        message_sender=MessageSender(address, port)
        message_sender.open_socket_and_send('LAMP_D2 1\n\r')
        sleep(1)
        message_sender.open_socket_and_send('LAMP_HAL 1\n\r')
        sleep(15)

    # define relevant strings
    def open_clarity_chrom(self, user: str,  config_file: str, password: str = None, start_method: str = ''):
        """
        start_method: supply the path to the method to start with, this is important for a soft column start
        config file: if you want to start with specific instrumment configuration, specify location of config file here
        """
        if not password:
            self.execute_command(f"i={self.instrument} cfg={config_file} u={user} {start_method}")
        else:
            self.execute_command(f"i={self.instrument} cfg={config_file} u={user} p={password} {start_method}")
        sleep(20)

    # TODO should be OS agnostic
    def slow_flowrate_ramp(self, path: str, method_list: tuple = ()):
        """
        path: path where the methods are located
        method list
        """
        for current_method in method_list:
            self.execute_command(f"i={self.instrument} {path}\\{current_method}")
            # not very elegant, but sending and setting method takes at least 10 seconds, only has to run during platform startup and can't see more elegant way how to do that
            sleep(20)


    def load_file(self, path_to_file: str):
        """has to be done to open project, then method. Take care to select 'Send Method to Instrument' option in Method
         Sending Options dialog in System Configuration."""
        self.execute_command(f"i={self.instrument} {path_to_file}")
        sleep(10)

    def set_sample_name(self, sample_name):
        """Sets the sample name for the next single run"""
        self.execute_command(f"i={self.instrument} set_sample_name={sample_name}")
        sleep(1)

    def run(self):
        """Runs the instrument. Care should be taken to activate automatic data export on HPLC. (can be done via command,
         but that only makes it more complicated). Takes at least 2 sec until run starts"""
        self.execute_command(f'run={self.instrument}')

    def exit(self):
        """Exit Clarity Chrom"""
        self.execute_command('exit')
        sleep(10)



class MessageSender:
    def __init__(self, host, port):
        self.host = host
        self.port = port

    # encode('utf-8')

    @tenacity.retry(stop=tenacity.stop_after_attempt(5), wait=tenacity.wait_fixed(2), reraise=True)
    def open_socket_and_send(self, message: str):
        s = socket.socket()
        s.connect((self.host, self.port))
        s.sendall(message.encode('utf-8'))
        s.close()


class ClarityExecutioner:
    """ This needs to run on the computer having claritychrom installed, except for one uses the same PC. However,
    going via socket and localhost would also work, but seems a bit cumbersome.
    open up server socket. Everything coming in will be prepended with claritychrom.exe (if it is not already)"""
    command_prepend = 'claritychrom.exe'

    def __init__(self, port, allowed_client='192.168.10.20', host_ip='192.168.10.11'):
        self.port = port
        self.allowed_client = allowed_client
        self.host_ip = host_ip
        # think that should also go in thread, otherwise blocks
        self.server_socket = self.open_server()
        self.executioner = Thread(target=self.get_commands_and_execute, daemon=False)
        print('a')
        self.executioner.start()
        print('b')

    def open_server(self):
        s = socket.socket()
        s.bind((self.host_ip, self.port))
        s.listen(5)
        return s

    def accept_new_connection(self):
        client_socket, address = self.server_socket.accept()
        if not address[0] == self.allowed_client:
            client_socket.close()
            print(f'nice try {client_socket, address}')
        else:
            # if below code is executed, that means the sender is connected
            print(f"[+] {address} is connected.")
            # in unicode
            request = client_socket.recv(1024).decode('utf-8')
            client_socket.close()
            print(request)
            return request


    # TODO: instrument number has to go into command execution
    def execute_command(self, command: str, folder_of_executable: Union[Path, str] = r'C:\claritychrom\bin\\'):
        prefix = 'claritychrom.exe'
        # sanitize input a bit
        if command.split(' ')[0] != prefix:
            command = folder_of_executable + prefix + ' ' + command  # type:ignore
            print(command)
        try:
            x = subprocess
            x.run(command, shell=True, capture_output=False, timeout=3)
        except subprocess.TimeoutExpired:
            print('Damn, Subprocess')

    def get_commands_and_execute(self):
        while True:
            request=self.accept_new_connection()
            self.execute_command(request)
            sleep(1)
            print('listening')

###TODO: also dsk or k for opening with specific desktop could be helpful-.
# TODO Export results can be specified -> exports result, rewrite to a nicer interface

if __name__ == "__main__":
    computer_w_Clarity = False
    if computer_w_Clarity  == True:
        analyser = ClarityExecutioner(10014)
    elif computer_w_Clarity == False:
        commander = ClarityInterface(remote=True, host='192.168.10.11', port=10014, instrument_number=2)
        commander.exit()
        commander.switch_lamp_on()  # address and port hardcoded
        commander.open_clarity_chrom("admin", config_file=r"C:\ClarityChrom\Cfg\automated_exp.cfg ", start_method=r"D:\Data2q\sugar-optimizer\autostartup_analysis\autostartup_005_Sugar-c18_shortened.MET")
        commander.slow_flowrate_ramp(r"D:\Data2q\sugar-optimizer\autostartup_analysis",
                                     method_list=("autostartup_005_Sugar-c18_shortened.MET",
                                                   "autostartup_01_Sugar-c18_shortened.MET",
                                                   "autostartup_015_Sugar-c18_shortened.MET",
                                                   "autostartup_02_Sugar-c18_shortened.MET",
                                                   "autostartup_025_Sugar-c18_shortened.MET",
                                                   "autostartup_03_Sugar-c18_shortened.MET",
                                                   "autostartup_035_Sugar-c18_shortened.MET",
                                                   "autostartup_04_Sugar-c18_shortened.MET",
                                                   "autostartup_045_Sugar-c18_shortened.MET",
                                                   "autostartup_05_Sugar-c18_shortened.MET",))
        commander.load_file(r"D:\Data2q\sugar-optimizer\autostartup_analysis\auto_Sugar-c18_shortened.MET")
        # commander.load_file("opendedicatedproject") # open a project for measurements
        commander.set_sample_name("test123")
        commander.run()



""" Functions related to the construction of instrument request """

from pathlib import WindowsPath

from lxml import etree


def create_message(sub_element_name, attributes=None):
    """
    Create a minimal XML tree with Message as root and sub_element as child tag
    """
    if attributes is None:
        attributes = {}

    root = etree.Element("Message")
    etree.SubElement(root, sub_element_name, attributes)
    return root


def set_attribute(name, value="") -> etree.Element:
    """
    Creates a Set message.
    Used for name = {Solvent | Sample} + indirectly by UserData and DataFolder
    """
    base = create_message("Set")
    attribute = etree.SubElement(base.find("./Set"), name)
    attribute.text = value
    return base


def get_request(name) -> etree.Element:
    """
    Creates a Get message.
    Used for name = {Solvent | Sample | UserData} + indirectly by UserData and DataFolder
    """
    base = create_message("GetRequest")
    etree.SubElement(base.find("./GetRequest"), name)
    return base


def set_data_folder(location) -> etree.Element:
    """
    Create a Set DataFolder message
    """
    # Get base request
    data_folder = set_attribute("DataFolder")

    # Add folder specific tag
    full_tree = etree.SubElement(data_folder.find(".//DataFolder"), "TimeStampTree")
    full_tree.text = (
        location.as_posix() if isinstance(location, WindowsPath) else location
    )

    return data_folder


def set_user_data(data: dict) -> etree.Element:
    """
    Given a dict with custom data, it creates a Set/UserData message.
    Those data are saved in acq.par
    """
    user_data = set_attribute("UserData")
    for key, value in data.items():
        etree.SubElement(
            user_data.find(".//UserData"), "Data", dict(key=key, value=value)
        )
    return user_data


def create_protocol_message(
    protocol_name: str, protocol_options: dict
) -> etree.Element:
    """
    Create an XML request to run a protocol
    """
    xml_root = create_message("Start", {"protocol": protocol_name})

    start_tag = xml_root.find("Start")

    for key, value in protocol_options.items():
        # All options are SubElements of the Start tag!
        etree.SubElement(start_tag, "Option", {"name": f"{key}", "value": f"{value}"})

    return xml_root



class Interface:
    """
    Represents *logical* graph edges, with Tube being its physical counterpart.
    """

    def __init__(self):
        ...

    def __repr__(self):
        return "Logical interface."



"""
Original code from Manson website with edits.
No license originally specified.
"""

import re
import warnings
from loguru import logger
from typing import Literal, Tuple, List, Union

import aioserial

from flowchem.exceptions import InvalidConfiguration, DeviceError
from flowchem.units import flowchem_ureg


class MansonPowerSupply:
    """Control module for Manson Power Supply (e.g. used to power LEDs in the photo-rector or as potentiostat)"""

    MODEL_ALT_RANGE = ["HCS-3102", "HCS-3014", "HCS-3204", "HCS-3202"]

    def __init__(self, port, baudrate=9600, **kwargs):
        try:
            self._serial = aioserial.Serial(
                port, baudrate=baudrate, timeout=0.1, **kwargs
            )  # type: aioserial.Serial
        except aioserial.SerialException as e:
            raise InvalidConfiguration(
                f"Could not connect to power supply on port <{port}>"
            ) from e

    def initialize(self):
        """Ensure the connection w/ device is working."""
        model_info = await self.get_info()
        if model_info == "":
            raise DeviceError("Communication with device failed!")
        if await self.get_info() not in self.MODEL_ALT_RANGE:
            raise InvalidConfiguration(
                f"Device is not supported! [Supported models: {self.MODEL_ALT_RANGE}]"
            )

    @staticmethod
    def _format_voltage(voltage_value: str) -> str:
        """ Format a voltage in the format the power supply understands """

        voltage = flowchem_ureg(voltage_value)
        # Zero fill by left pad with zeros, up to three digits
        return str(voltage.m_as("V") * 10).zfill(3)

    async def _format_amperage(self, amperage_value: str) -> str:
        """ Format a current intensity in the format the power supply understands """

        current = flowchem_ureg(amperage_value)
        multiplier = 100 if await self.get_info() in self.MODEL_ALT_RANGE else 10
        return str(current.m_as("A") * multiplier).zfill(3)

    async def _send_command(
        self,
        command: str,
    ) -> str:
        """Internal function to send command and read reply."""

        # Flush buffer
        self._serial.reset_input_buffer()

        # Write command
        await self._serial.write_async(f"{command}\r".encode("ascii"))

        # Read reply
        reply_string = []
        for line in await self._serial.readlines_async():
            reply_string.append(line.decode("ascii").strip())
            logger.debug(f"Received {repr(line)}!")

        return "\n".join(reply_string)

    async def get_info(self) -> str:
        """Returns the model name of the connected device"""
        response = await self._send_command("GMOD")

        pattern = re.compile(r".*\d{4}\s")
        match = pattern.match(response)

        if match:
            if response[0:4] == "HCS-":
                return match.group().rstrip()
            return "HCS-" + match.group().rstrip()
        return ""

    async def output_on(self) -> bool:
        """Turn on electricity on output"""
        response = await self._send_command("SOUT0")
        return response == "OK"

    async def output_off(self) -> bool:
        """Turn off electricity on output"""
        response = await self._send_command("SOUT1")
        return response == "OK"

    async def get_output_read(
        self,
    ) -> Tuple[str, str, Union[Literal["CC"], Literal["CV"], Literal["NN"]]]:
        """Returns actual values of voltage, current and mode"""
        response = await self._send_command("GETD")

        try:
            volt = float(response[0:4]) / 100 * flowchem_ureg.volt
            curr = float(response[4:8]) / 100 * flowchem_ureg.ampere
        except ValueError:
            warnings.warn("Invalid values from device!")
            return "0 V", "0 A", "NN"

        if response[8:9] == "0":
            return str(volt), str(curr), "CV"
        elif response[8:9] == "1":
            return str(volt), str(curr), "CC"
        else:
            return str(volt), str(curr), "NN"

    async def get_output_voltage(self) -> str:
        """Returns output voltage in Volt"""
        voltage, _, _ = await self.get_output_read()
        return voltage

    async def get_output_current(self) -> str:
        """Returns output current in Ampere"""
        _, current, _ = await self.get_output_read()
        return current

    async def get_output_mode(self) -> Literal["CC", "CV", "NN"]:
        """Returns output mode: either current control (CC) or voltage control (CV)"""
        _, _, mode = await self.get_output_read()
        return mode

    async def get_output_power(self) -> str:
        """Returns output power in watts"""
        voltage, intensity, _ = await self.get_output_read()
        power = flowchem_ureg(voltage) * flowchem_ureg(intensity)
        return str(power.to("W"))

    async def get_max(self) -> Tuple[str, str]:
        """Returns maximum voltage and current, as tuple, or False."""
        response = await self._send_command("GMAX")

        max_v_raw = int(response[0:3]) * flowchem_ureg.volt
        max_c_raw = int(response[3:6]) * flowchem_ureg.ampere

        max_v = max_v_raw / 10
        # Some models report current as 0.1 A others at 0.01 A
        model = await self.get_info()
        divider = 100 if model in self.MODEL_ALT_RANGE else 10
        return str(max_v), str(max_c_raw / divider)

    async def get_setting(self) -> Tuple[str, str]:
        """Returns current setting as tuple (voltage, current)."""
        response = await self._send_command("GETS")

        # RegEx to only keep numbers
        response = re.sub(r"\D", "", response)
        v_setting = float(response[0:3]) / 10 * flowchem_ureg.volt
        c_setting = float(response[3:6]) * flowchem_ureg.ampere

        if await self.get_info() in self.MODEL_ALT_RANGE:
            c_setting /= 10

        return str(v_setting), str(c_setting / 10)

    async def set_voltage(self, voltage: str) -> bool:
        """Set target voltage"""

        cmd = "VOLT" + self._format_voltage(voltage)
        response = await self._send_command(cmd)
        return response == "OK"

    async def set_current(self, current: str) -> bool:
        """Set target current"""

        cmd = "CURR" + await self._format_amperage(current)
        response = await self._send_command(cmd)
        return response == "OK"

    async def set_all_preset(self, preset: List[Tuple[str, str]]) -> bool:
        """Set all 3 preset memory position with voltage/current values"""
        command = "PROM"

        for set_values in preset:
            voltage, current = set_values
            voltage_string = self._format_voltage(voltage)
            current_string = await self._format_amperage(current)
            command += voltage_string + current_string

        # Set new values (no reply from device on this command)
        await self._send_command(command)
        return True

    async def set_preset(self, index: int, voltage: str, current: str) -> bool:
        """Set preset position index with the provided values of voltage and current"""
        preset = await self.get_all_preset()
        try:
            volt_str = self._format_voltage(voltage)
            curr_str = await self._format_amperage(current)
            preset[index] = (volt_str, curr_str)
        except KeyError:
            warnings.warn(f"Preset {index} not found! Command ignored")
            return False
        return await self.set_all_preset(preset)

    async def get_all_preset(self) -> List[Tuple[str, str]]:
        """Get voltage and current for all 3 memory preset position"""
        response = await self._send_command("GETM")
        response_lines = response.split("\r")

        voltage = []
        current = []

        for preset in response_lines[0:3]:
            # Drop all but numbers
            preset = re.sub(r"\D", "", preset)
            try:
                # Three digits for voltage and three digits for current
                voltage.append(float(preset[0:3]))
                current.append(float(preset[3:6]))
            except (KeyError, ValueError):
                warnings.warn("Error reading presets!")

        # Transform current in Ampere and voltage in Volt
        current = [x / 10 for x in current]
        voltage = [x / 10 for x in voltage]

        # Usual issue with some models having higher precision for voltage
        if await self.get_info() in self.MODEL_ALT_RANGE:
            voltage = [x / 10 for x in voltage]

        voltage_str = [str(flowchem_ureg.Quantity(x, "V")) for x in voltage]
        current_str = [str(flowchem_ureg.Quantity(x, "A")) for x in current]

        return list(zip(voltage_str, current_str))

    async def get_preset(self, index) -> Tuple[str, str]:
        """Get voltage and current for given preset position [0..2]"""
        all_preset = await self.get_all_preset()
        try:
            return all_preset[index]
        except KeyError:
            warnings.warn(f"Preset {index} not found! Command ignored")
            return "", ""

    async def run_preset(self, index: int) -> bool:
        """Set Voltage and Current using values saved in one of the three memory locations: 0, 1 or 2"""
        if not 0 <= int(index) < 3:
            warnings.warn(f"Invalid preset value: <{index}>!")
            return False
        cmd = "RUNM" + str(int(index))
        response = await self._send_command(cmd)
        return response == "OK"

    async def remove_protection(self) -> bool:
        """ " I guess it removes over voltage protection?"""
        response = await self._send_command("SPRO0")
        return bool(response)

    async def add_protection(self) -> bool:
        """ " I guess it adds over voltage protection?"""
        response = await self._send_command("SPRO1")
        return bool(response)

    async def set_voltage_and_current(
        self, voltage: str, current: str
    ):
        """ Convenience method to set both voltage and current """
        await self.set_voltage(voltage)
        await self.set_current(current)

    def get_router(self):
        """Creates an APIRouter for this MansonPowerSupply instance."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/output/on", self.output_on, methods=["GET"])
        router.add_api_route("/output/off", self.output_off, methods=["GET"])
        router.add_api_route("/output/power", self.get_output_power, methods=["GET"])
        router.add_api_route("/output/mode", self.get_output_mode, methods=["GET"])
        router.add_api_route("/voltage/read", self.get_output_voltage, methods=["GET"])
        router.add_api_route("/voltage/max", self.set_voltage, methods=["PUT"])
        router.add_api_route("/current/read", self.get_output_current, methods=["GET"])
        router.add_api_route("/current/max", self.set_current, methods=["PUT"])
        router.add_api_route("/protection/add", self.add_protection, methods=["GET"])
        router.add_api_route(
            "/protection/remove", self.remove_protection, methods=["GET"]
        )

        return router



from flowchem.components.stdlib import Component


class MappedComponentMixin(Component):
    """
    A Mixin to be added to component with mapping.

    For more details on Mixins, see Fluen Python chapter 14 or Effective Python item 41 (links are free for MPG):
    - https://learning.oreilly.com/library/view/fluent-python-2nd/9781492056348/ch14.html#idm45517018812200
    - https://learning.oreilly.com/library/view/effective-python-90/9780134854717/ch05.xhtml#item41

    All components with mapping should derive from this, both active (e.g. valves) and passive (reactors).

    Arguments:
    - `name`: The name of the component.

    """

    def _validate(self, dry_run):
        if not self.mapping:
            raise ValueError(f"{self} requires a mapping. None provided.")
        return super()._validate(dry_run)



from __future__ import annotations

import inspect
import itertools
import json
import os
from loguru import logger
from collections import namedtuple
from pathlib import Path
from types import ModuleType
from typing import Iterable, Dict, Optional, Any, List, Union

import jsonschema
import yaml

import flowchem.components.devices
from flowchem.components.stdlib import Tube, Interface
from flowchem.core.apparatus import Apparatus
from flowchem.core.graph.DeviceNode import DeviceNode
from flowchem.exceptions import InvalidConfiguration

# packages containing the device class definitions. Target classes should be available in the module top level.
DEVICE_MODULES = [flowchem.components.devices, flowchem.components.stdlib, flowchem.components.reactors]

# Validation schema for graph file
SCHEMA = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "../graph/flowchem-graph-spec.schema"
)


def get_device_class_mapper(modules: Iterable[ModuleType]) -> Dict[str, type]:
    """
    Given an iterable of modules containing the device classes, return a
    dictionary Dict[device_class_name, DeviceClass]

    Args:
        modules (Iterable[ModuleType]): The modules to inspect for devices.
            Only class in the top level of each module will be extracted.
    Returns:
        device_dict (Dict[str, type]): Dict of device class names and their
            respective classes, i.e. {device_class_name: DeviceClass}.
    """
    # Get (name, obj) tuple for the top level of each modules.
    objects_in_modules = [
        inspect.getmembers(module, inspect.isclass) for module in modules
    ]

    # Return them as dict (itertools to flatten the nested, per module, lists)
    return {k: v for (k, v) in itertools.chain.from_iterable(objects_in_modules)}


def load_schema():
    """loads the schema defining valid config file."""
    with open(SCHEMA, "r") as fp:
        schema = json.load(fp)
        jsonschema.Draft7Validator.check_schema(schema)
        return schema


Connection = namedtuple("Connection", ["from_device", "to_device", "tube"])


class DeviceGraph:
    """
    Represents the device graph.

    This borrows logic from mw.Apparatus and ChempilerGraph
    """

    _id_counter = 0

    def __init__(self, configuration, name: Optional[str] = None):
        # if given a name, then name the apparatus, else default to a sequential name
        if name is not None:
            self.name = name
        else:
            self.name = "DeviceGraph" + str(DeviceGraph._id_counter)
            DeviceGraph._id_counter += 1

        # dict of devices with names as keys
        self.device: Dict[str, Any] = {}
        # Edge list represents the network topology
        self.edge_list: List[Connection] = []

        # Save config pre-parsing for debug purposes
        self._raw_config = configuration

        # Load graph
        # self.validate(configuration)
        self.parse(configuration)

    @classmethod
    def from_file(cls, file: Union[Path, str]):
        """Creates DeviceGraph from config file"""

        file_path = Path(file)
        name = file_path.stem

        with file_path.open() as stream:
            config = yaml.safe_load(stream)

        return cls(config, name)

    def validate(self, config):
        """Validates config syntax."""
        schema = load_schema()
        jsonschema.validate(config, schema=schema)

    def parse(self, config: Dict):
        """Parse config and generate graph."""

        # Device mapper
        device_mapper = get_device_class_mapper(DEVICE_MODULES)
        logger.debug(
            f"The following device classes have been found: {device_mapper.keys()}"
        )

        # Parse list of devices and create nodes
        for device_name, node_config in config["devices"].items():
            # Schema validation should ensure 1 hit here
            try:
                device_class = [
                    name for name in device_mapper.keys() if name in node_config
                ].pop()
            except IndexError as e:
                raise InvalidConfiguration(f"Node config invalid: {node_config}") from e

            # Object type
            obj_type = device_mapper[device_class]
            device_config = node_config[device_class]

            self.device[device_name] = DeviceNode(
                device_name, device_config, obj_type
            ).device
            logger.debug(
                f"Created device <{device_name}> with config: {device_config}"
            )

        # Parse list of connections
        connections = self._parse_connections(config["connections"])
        self.edge_list.extend(connections)

    def _parse_connections(self, connections: Dict) -> List[Connection]:
        """ Parse connections from config. """
        connection_list = []
        for edge in connections:
            if "Tube" in edge:
                connection = self.parse_tube_connection(edge["Tube"])
            elif "Interface" in edge:
                connection = self.parse_interface_connection(edge["Interface"])
            else:
                raise InvalidConfiguration(f"Invalid connection type in {edge}")

            connection_list.append(connection)

        return connection_list

    def parse_tube_connection(self, tube_config) -> Connection:
        """ Parse a dict containing the Tube connection and returns the Connection """
        tube = Tube(
            length=tube_config["length"],
            ID=tube_config["inner-diameter"],
            OD=tube_config["outer-diameter"],
            material=tube_config["material"],
        )

        # Devices
        try:
            from_device = self.device[tube_config["from"]["device"]]
            to_device = self.device[tube_config["to"]["device"]]
        except KeyError as ke:
            raise InvalidConfiguration("A Tube refers to a non existing node!\n"
                                       f"Missing node: {ke}\n"
                                       f"Tube config: {tube_config}") from ke

        # If necessary updates mapping.
        if tube_config["from"].get("position", 0) != 0:
            from_device.mapping[from_device.name] = tube_config["from"]["position"]
        if tube_config["to"].get("position", 0) != 0:
            to_device.mapping[from_device.name] = tube_config["to"]["position"]

        return Connection(from_device, to_device, tube)

    def parse_interface_connection(self, iface_config) -> Connection:
        """ Parse a dict containing the Tube connection and returns the Connection """
        interface = Interface()

        try:
            from_device = self.device[iface_config["from"]["device"]]
            to_device = self.device[iface_config["to"]["device"]]
        except KeyError as ke:
            raise InvalidConfiguration("An Interface refers to a non existing node!\n"
                                       f"Missing node: {ke}\n"
                                       f"Interface config: {iface_config}") from ke

        return Connection(from_device, to_device, interface)

    def to_apparatus(self) -> Apparatus:
        """
        Convert the graph to an mw.Apparatus object.
        """

        appa = Apparatus(
            self.name, "Apparatus auto-generated from flowchem DeviceGraph."
        )
        for edge in self.edge_list:
            print(edge)
            print(edge.from_device)
            print(edge.to_device)
            print(edge.tube)

            appa.add(edge.from_device, edge.to_device, edge.tube)
        return appa

    def __repr__(self):
        return f"<DeviceGraph {self.name}>"

    def __str__(self):
        return f"DeviceGraph {self.name} with {len(self.device)} devices."

    def __getitem__(self, item):
        """
        Utility method

        DeviceGraph['name'] gives the device with that name
        DeviceGraph[class] returns a list of devices of that type
        DeviceGraph[device_instance] returns true if the object is part of the graph
        """

        # If a type is passed return devices with that type
        if isinstance(item, type):
            return [
                device for device in self.device.values() if isinstance(device, item)
            ]
        # If a string is passed return the device with that name
        elif isinstance(item, str):
            try:
                return self.device[item]
            except IndexError:
                raise KeyError(f"No component named '{item}' in {repr(self)}.")

        # a shorthand way to check if a component is in the apparatus
        elif item in self.device.values():
            return item
        else:
            raise KeyError(f"{repr(item)} is not in {repr(self)}.")


if __name__ == "__main__":
    from flowchem import Protocol
    from datetime import timedelta

    graph = DeviceGraph.from_file("owen_config.yml")

    a = graph.to_apparatus()
    print(a)
    p = Protocol(a)

    t0 = timedelta(seconds=0)

    # p.add(graph["quencher"], start=t0, duration=timedelta(seconds=10), rate="0.1 ml/min")
    p.add(graph["activator"], start=t0, duration=timedelta(seconds=10), rate="0.1 ml/min")
    print(graph["chiller"])
    print(type(graph["chiller"]))
    p.add(graph["chiller"], start=t0, duration=timedelta(seconds=10), temp="45 degC")

    E = p.execute(dry_run=False)
    # E.visualize()



"""
This module is used to control Vici Valco Universal Electronic Actuators.
"""

from __future__ import annotations

from loguru import logger
from dataclasses import dataclass
from typing import Optional, Set

import aioserial

from flowchem.components.stdlib import Valve
from flowchem.exceptions import InvalidConfiguration, ActuationError, DeviceError


@dataclass
class ViciProtocolCommandTemplate:
    """Class representing a valve command and its expected reply, but without target valve number"""

    command: str
    optional_parameter: str = ""

    def to_valve(
        self, address: int, command_value: str = "", argument_value: str = ""
    ) -> ViciProtocolCommand:
        """Returns a Protocol11Command by adding to the template valve address and command arguments"""
        return ViciProtocolCommand(
            target_valve_num=address,
            command=self.command,
            optional_parameter=self.optional_parameter,
            command_value=command_value,
            argument_value=argument_value,
        )


@dataclass
class ViciProtocolCommand(ViciProtocolCommandTemplate):
    """Class representing a valve command and its expected reply"""

    target_valve_num: Optional[int] = 1
    command_value: Optional[str] = None
    argument_value: Optional[str] = None

    def compile(self) -> bytes:
        """Create actual command byte by prepending valve address to command and appending executing command."""

        assert self.target_valve_num in range(0, 11)
        if not self.command_value:
            self.command_value = ""

        compiled_command = (
            f"{self.target_valve_num}" f"{self.command}{self.command_value}"
        )

        if self.argument_value:
            compiled_command += f"{self.optional_parameter}{self.argument_value}"

        return (compiled_command + "\r").encode("ascii")


class ViciValcoValveIO:
    """Setup with serial parameters, low level IO"""

    DEFAULT_CONFIG = {
        "timeout": 0.5,
        "baudrate": 9600,
        "parity": aioserial.PARITY_NONE,
        "stopbits": aioserial.STOPBITS_ONE,
        "bytesize": aioserial.EIGHTBITS,
    }

    def __init__(self, aio_port: aioserial.Serial):
        """
        Initialize communication on the serial port where the valves are located and initialize them
        Args:
            aio_port: aioserial.Serial() object
        """

        self._serial = aio_port

        # These will be set in initialize
        self.num_valve_connected = 0
        self._initialized = False

    @classmethod
    def from_config(cls, port, **serial_kwargs):
        """Create ViciValcoValveIO from config."""
        # Merge default serial settings with provided ones.
        configuration = dict(ViciValcoValveIO.DEFAULT_CONFIG, **serial_kwargs)

        try:
            serial_object = aioserial.AioSerial(port, **configuration)
        except aioserial.SerialException as e:
            raise InvalidConfiguration(
                f"Cannot connect to the valve on the port <{port}>"
            ) from e

        return cls(serial_object)

    async def initialize(self, hw_initialization: bool = True):
        """Ensure connection + initialize."""
        # This has to be run after each power cycle to assign addresses to valves
        self.num_valve_connected = await self.detect_valve_address()

        if hw_initialization:
            self._hw_init()

        self._initialized = True

    async def detect_valve_address(self) -> int:
        """Detects number of valves connected."""
        try:
            await self._serial.write_async("*ID\r".encode("ascii"))
        except aioserial.SerialException as e:
            raise InvalidConfiguration from e

        reply = self._serial.readlines()
        n_valves = len(reply)
        if n_valves == 0:
            raise InvalidConfiguration(f"No valve found on {self._serial.port}")
        else:
            logger.debug(f"Found {len(reply)} valves on {self._serial.port}!")
            return len(reply)

    def _hw_init(self):
        """Send to all valves the HW initialization command (i.e. homing)"""
        self._serial.write("*HM\r".encode("ascii"))  # Broadcast: initialize + execute
        # Note: no need to consume reply here because there is none (since we are using broadcast)

    async def _write(self, command: bytes):
        """Writes a command to the valve"""
        if not self._initialized:
            raise DeviceError(
                "Valve not initialized!\n"
                "Have you called `initialize()` after object creation?"
            )
        await self._serial.write_async(command)
        logger.debug(f"Command {repr(command)} sent!")

    async def _read_reply(self, lines) -> str:
        """Reads the valve reply from serial communication"""
        reply_string = ""
        for line in range(lines):
            a = b""
            a = await self._serial.readline_async()
            reply_string += a.decode("ascii")

        logger.debug(f"Reply received: {reply_string}")
        return reply_string

    def reset_buffer(self):
        """Reset input buffer before reading from serial. In theory not necessary if all replies are consumed..."""
        self._serial.reset_input_buffer()

    async def write_and_read_reply(self, command: ViciProtocolCommand, lines) -> str:
        """Main ViciValcoValveIO method.
        Sends a command to the valve, read the replies and returns it, optionally parsed"""
        self.reset_buffer()
        await self._write(command.compile())

        if not lines:
            return ""

        response = await self._read_reply(lines)

        if not response:
            raise InvalidConfiguration(
                f"No response received from valve, check valve address! "
                f"(Currently set to {command.target_valve_num})"
            )
        return response.rstrip()

    @property
    def name(self) -> str:
        """This is used to provide a nice-looking default name to valves based on their serial connection."""
        try:
            return self._serial.name
        except AttributeError:
            return ""


class ViciValco(Valve):
    """ " """

    # This class variable is used for daisy chains (i.e. multiple valves on the same serial connection). Details below.
    _io_instances: Set[ViciValcoValveIO] = set()
    # The mutable object (a set) as class variable creates a shared state across all the instances.
    # When several valves are daisy chained on the same serial port, they need to all access the same Serial object,
    # because access to the serial port is exclusive by definition (also locking there ensure thread safe operations).
    # FYI it is a borg idiom https://www.oreilly.com/library/view/python-cookbook/0596001673/ch05s23.html

    valve_position_name = {"A": 1, "B": 2}

    def __init__(self, valve_io: ViciValcoValveIO, address: int = 0, name: str = None):
        """
        Default constructor, needs an ViciValcoValveIO object. See from_config() class method for config-based init.
        Args:
            valve_io: An ViciValcoValveIO w/ serial connection to the daisy chain w/ target valve.
            address: number of valve in array, 1 for first one, auto-assigned on init based on position.
            name: 'cause naming stuff is important.
        """

        # ViciValcoValveIO
        self.valve_io = valve_io
        ViciValco._io_instances.add(self.valve_io)  # See above for details.

        # The valve name is used for logs and error messages.
        self.name = f"Valve {self.valve_io.name}:{address}" if name is None else name

        super().__init__(None, name)

        # valve address is the valve sequence number if in chain. Count starts at 1, default.
        self.address = int(address)

    @classmethod
    def from_config(cls, port: str, address: int, name: str = None, **serial_kwargs):
        """This class method is used to create instances via config file by the server for HTTP interface."""
        # Many valve can be present on the same serial port with different addresses.
        # This shared list of ViciValcoValveIO objects allow shared state in a borg-inspired way, avoiding singletons
        # This is only relevant to programmatic instantiation, i.e. when from_config() is called per each valve from a
        # config file, as it is the case in the HTTP server.
        # ViciValcoValve_IO() manually instantiated are not accounted for.
        valveio = None
        for obj in ViciValco._io_instances:
            if obj._serial.port == port:
                valveio = obj
                break

        # If not existing serial object are available for the port provided, create a new one
        if valveio is None:
            valveio = ViciValcoValveIO.from_config(port, **serial_kwargs)

        return cls(valveio, address=address, name=name)

    async def initialize(self):
        """Must be called after init before anything else."""
        # Test connectivity by querying the valve's firmware version
        fw_cmd = ViciProtocolCommandTemplate(command="VR").to_valve(self.address)
        firmware_version = await self.valve_io.write_and_read_reply(fw_cmd, lines=5)
        logger.info(
            f"Connected to Vici Valve {self.name} - FW version: {firmware_version}!"
        )

    async def send_command_and_read_reply(
        self,
        command_template: ViciProtocolCommandTemplate,
        command_value="",
        argument_value="",
        lines=1,
    ) -> str:
        """Sends a command based on its template by adding valve address and parameters, returns reply"""
        return await self.valve_io.write_and_read_reply(
            command_template.to_valve(self.address, command_value, argument_value),
            lines,
        )

    async def learn_valve_positions(self) -> None:
        """Initialize valve only, there is no reply -> lines = 0"""
        await self.send_command_and_read_reply(
            ViciProtocolCommandTemplate(command="LRN"), lines=0
        )

    async def initialize_valve(self) -> None:
        """Initialize valve only: Move to Home position"""
        await self.send_command_and_read_reply(
            ViciProtocolCommandTemplate(command="HM"), lines=0
        )
        # seems necessary to make sure move is finished
        await self.get_valve_position()

    async def version(self) -> str:
        """Returns the current firmware version reported by the valve."""

        return await self.send_command_and_read_reply(
            ViciProtocolCommandTemplate(command="VR"), lines=5
        )

    async def get_valve_position(self) -> int:
        """Represent the position of the valve: getter returns Enum, setter needs Enum."""
        valve_pos = await self.send_command_and_read_reply(
            ViciProtocolCommandTemplate(command="CP")
        )
        return ViciValco.valve_position_name[valve_pos[-1]]

    async def set_valve_position(self, target_position: int):
        """Set valve position. Switches really quick and doesn't reply, so waiting does not make sense."""
        valve_by_name_cw = ViciProtocolCommandTemplate(command="GO")
        await self.send_command_and_read_reply(
            valve_by_name_cw, command_value=str(target_position), lines=0
        )
        logger.debug(f"{self.name} valve position set to {target_position}")
        new_position = await self.get_valve_position()
        if not new_position == target_position:
            raise ActuationError

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    async def _update(self):
        """ Used in automation. """
        await self.set_valve_position(self.setting)

    def get_router(self):
        """Creates an APIRouter for this object."""
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route("/firmware-version", self.version, methods=["GET"])
        router.add_api_route(
            "/initialize/valve", self.initialize_valve, methods=["PUT"]
        )
        router.add_api_route(
            "/valve/position", self.get_valve_position, methods=["GET"]
        )
        router.add_api_route(
            "/valve/position", self.set_valve_position, methods=["PUT"]
        )

        return router


if __name__ == "__main__":
    import asyncio

    valve1 = ViciValco.from_config(port="COM13", address=0, name="test1")

    asyncio.run(valve1.initialize_valve())

    asyncio.run(valve1.set_valve_position(2))

    asyncio.run(valve1.set_valve_position(1))


# Control Command List for reference, don't see much of a point to implement all these,
# especially since most don't return anything
#
# GO[nn]     - Move to nn position -> None
#
# HM         - Move to the first Position -> None
#
# CW[nn]     - Move Clockwise to nn Position ->
#
# CC[nn]     - Move Counter Clockwise to nn Position ->
#
# TO         - Toggle Position to Oposite ->
#
# TT         - Timed Toggle
#
# DT[nnnnn]  - Set Delay time for TT Command
#
# CP         - Returns Current Position -> [ADDRESS]CP[A|B]
#
# AM[n]      - Sets the Actuator Mode [1] Two Position With Stops, -> [ADDRESS]AM[1|2|3]
#
#              [2] Two Position Without Stops, [3] Multi Position
#
# SB[nnnnn]  - Set the Baud Rate to nnnnn -> [ADDRESS]SB[BAUDRATE:int]
#
# NP[nn]     - Set the Number of Positions to nn -> 0E2 NP Invalid
#
# SM[n]      - Set the Direction [F]orward, [R]everse, [A]uto 0E2 SM Invalid
#
# LRN        - Learn Stops Location -> None
#
# CNT[nnnnn] - Set Cycle Counter -> 0CNT10254
#
# VR[n]      - Firmware Version [] Main [1] Display [2] Interface -> 0Dec 15 2011 \n 015:02:20 \n 0UA_MAIN_CT
#
# ID[nn]     - Set Device ID nn=(0-9, A-Z) -> 0ID0
#
# [n]ID*     - Reset ID to none n=Current ID
#
# IFM[n]     - Interface Mode [0] No Response [1] limited response -> 0IFM0
#
#              [2] Extended Response
#
# LG[n]      - Legacy Response Mode [0] Off [1] On -> 0LG0
#
# /?         - Displays This List



""" For each device node described in the configuration [graph] instantiated it and create endpoints """
import inspect
import warnings

from loguru import logger
from fastapi import APIRouter

from flowchem import Spinsolve
from flowchem.core.server.routers import spinsolve_get_router


class DeviceNode:
    """Represent a node in the device graph, holds the HW object and its metadata/config."""

    # Router generators for device class that do not implement self.get_router()
    # All callable take the device obj and return an APIRouter
    router_generator = {
        Spinsolve: spinsolve_get_router
    }

    def __init__(self, device_name, device_config, obj_type):
        self._title = device_name
        self._router = None

        # No configuration for t-mixer et al.
        if device_config is None:
            device_config = {}

        # Add name to initialization args
        if "name" not in device_config:
            device_config["name"] = device_name

        # DEVICE INSTANTIATION
        try:
            # Special class method for initialization required for some devices
            if hasattr(obj_type, "from_config"):
                self.device = obj_type.from_config(**device_config)
            else:
                self.device = obj_type(**device_config)
            logger.debug(f"Created {self.title} instance: {self.device}")
        except TypeError as e:
            raise ConnectionError(
                f"Wrong configuration provided for device: {self.title} of type {obj_type}!\n"
                f"Configuration: {device_config}\n"
                f"Accepted parameters: {inspect.getfullargspec(obj_type).args}"
            ) from e

        # self.service_info = None

    @property
    def router(self):
        """Returns an APIRouter associated with the device"""
        if self._router:
            return self._router

        if hasattr(self.device, "get_router"):
            router = self.device.get_router()
        else:
            try:
                router = DeviceNode.router_generator[type(self.device)](self.device)
            except KeyError:
                warnings.warn(f"No router available for device '{self.device.name}'")
                router = APIRouter()

        router.prefix = f"/{self.safe_title}"
        router.tags = [self.safe_title]
        self._router = router
        return self._router

    # @property
    # def description(self) -> str:
    #     """ Human-readable description of the Node """
    #     return self._description
    #
    # @description.setter
    # def description(self, description: str):
    #     """
    #     Human-readable description of the Node
    #     :param description: str:
    #     """
    #     self._description = description

    @property
    def title(self) -> str:
        """Human-readable title of the Node"""
        return self._title

    @title.setter
    def title(self, title: str):
        """
        Human-readable title of the Node
        :param title: str:
        """
        self._title = title

    @property
    def safe_title(self) -> str:
        """Lowercase title with no whitespace"""
        title = self.title
        if not title:
            title = "unknown"
        title = title.replace(" ", "")
        title = title.lower()
        return title



from typing import Optional

from flowchem.units import flowchem_ureg
from flowchem.components.stdlib import ActiveComponent


class TempControl(ActiveComponent):
    """
    A generic temperature controller.

    Arguments:
    - `internal_tubing`: The `Tube` inside the temperature controller.
    - `name`: The component's name.

    Attributes:
    - `active`: Whether the temperature controller is active.
    - `name`: The name of the Sensor.
    - `temp`: The temperature setting as a `pint.Quantity`.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)

        self.temp = flowchem_ureg.parse_expression("0 degC")
        self.active = False

        self._base_state = dict(temp="0 degC", active=False)

    async def _update(self):
        raise NotImplementedError



from typing import Mapping, Optional

from flowchem.components.stdlib import ActiveComponent, Component


class Valve(ActiveComponent):
    """
    A generic valve.

    Arguments:
    - `mapping`: The mapping from components to their integer port numbers.
    - `name`: The name of the valve.

    Attributes:
    - `mapping`: The mapping from components to their integer port numbers.
    - `name`: The name of the valve.
    - `setting`: The position of the valve as an int (mapped via `mapping`).
    """

    def __init__(
        self,
        mapping: Optional[Mapping[Component, int]] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)

        # check the mapping's type
        if not isinstance(mapping, (type(None), Mapping)):
            raise TypeError(f"Invalid mapping type {type(mapping)} for {repr(self)}.")
        self.mapping = mapping
        self.setting = 1
        self._visualization_shape = "parallelogram"

        self._base_state = {"setting": 1}

    async def _update(self):
        raise NotImplementedError

    def _validate(self, dry_run):
        if not self.mapping:
            raise ValueError(f"{self} requires a mapping. None provided.")
        return super()._validate(dry_run)



""" LTF reactors """
from typing import Optional, Mapping

from flowchem.components.stdlib import Component, MappedComponentMixin


class LTF_HTM_ST_3_1(MappedComponentMixin, Component):
    """
    An LTF HTM ST 3 1 reactor.
    """

    def __init__(self, mapping: Optional[Mapping[Component, int]] = None, name: Optional[str] = None):
        super().__init__(name=name)
        if not isinstance(mapping, (type(None), Mapping)):
            raise TypeError(f"Invalid mapping type {type(mapping)} for {repr(self)}.")
        self.mapping = mapping
        self._visualization_shape = "cds"

    def _validate(self, dry_run):
        if not self.mapping:
            raise ValueError(f"{self} requires a mapping. None provided.")
        return super()._validate(dry_run)



from typing import Optional

from flowchem.units import flowchem_ureg
from flowchem.components.stdlib import ActiveComponent


class Pump(ActiveComponent):
    """
    A generic pumping device whose primary feature is that it moves fluid.

    Arguments:
    - `name`: The name of the pump.

    Attributes:
    - `name`: The name of the pump.
    - `rate`: The flow rate of the pump as a `pint.Quantity`. Must be of the dimensionality of volume/time.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self.rate = flowchem_ureg.parse_expression("0 ml/min")
        self._visualization_shape = "box3d"
        self._base_state = dict(rate="0 mL/min")



""" Creates mDNS server """
import hashlib
import logging
import uuid

# noinspection PyProtectedMember
from zeroconf import IPVersion, ServiceInfo, Zeroconf, get_all_addresses


class Server_mDNS:
    """mDNS server.
    :param host: Host IP address. Defaults to 0.0.0.0.
    :type host: string
    :param port: Host port. Defaults to 7485.
    :type port: int
    :param debug: Enable server debug mode. Defaults to False.
    :type debug: bool
    """

    def __init__(self, host="0.0.0.0", port=7485, debug=False):
        # Server properties
        self.host = host
        self.port = port
        self.debug = debug

        self.server = Zeroconf(ip_version=IPVersion.V4Only)

        # Get list of host addresses
        self.mdns_addresses = [
            ip
            for ip in get_all_addresses()  # Get all local IP
            if ip not in ("127.0.0.1", "0.0.0.0")
            and not ip.startswith("169.254")  # Remove invalid IPs
        ]

        # Logger
        self.logger = logging.getLogger(__name__)

    @staticmethod
    def _get_valid_service_name(name: str):
        """Given a desired service name, returns a valid one ;)"""
        candidate_name = f"{name}._labthing._tcp.local."
        if len(candidate_name) < 64:
            return candidate_name
        else:
            # Hexdigest and not trimming to ensure uniqueness
            return f"{hashlib.sha1(candidate_name.encode()).hexdigest()}._labthing._tcp.local."

    def include_device(self, name, url_prefix):
        """Adds device to the server."""
        service_name = Server_mDNS._get_valid_service_name(name)

        # LabThing service
        service_info = ServiceInfo(
            type_="_labthing._tcp.local.",
            name=service_name,
            port=self.port,
            properties={
                "path": url_prefix,
                "id": f"{service_name}:{uuid.uuid4()}".replace(" ", ""),
            },
            parsed_addresses=self.mdns_addresses,
        )

        self.server.register_service(service_info)
        self.logger.debug(f"Registered {service_name} on mDNS server!")


if __name__ == "__main__":
    test = Server_mDNS()
    print("ok")
    input()



""" Util function to install dependencies on PC with no internet access. """
import argparse
from pathlib import Path
import subprocess
import sys

EXCHANGE_FOLDER = Path(r"W:\BS-FlowChemistry\Resources\python_packages_local")


def install_from_folder(package: str, folder: Path):
    """pip-install packages locally available in a folder."""
    subprocess.check_call(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--no-index",
            f"--find-links={folder.as_posix()}",
            package,
        ]
    )


def download_to_folder(package: str, folder: Path):
    """pip-download packages to a local folder."""
    subprocess.check_call(
        [sys.executable, "-m", "pip", "download", package, "-d", folder.as_posix()]
    )


def get_package_list():
    """Return the list of packages to download/install from the requirements file"""
    req_file = Path("../../requirements.txt")

    package = []

    with req_file.open() as fh:
        lines = fh.readlines()
        for line in lines:
            # Ignore nmrglue
            if "nmrglue" in line:
                continue

            package.append(line.split("~=")[0])

    return package


def download_all():
    """Downloads all packages in requirements."""
    for package in get_package_list():
        download_to_folder(package, EXCHANGE_FOLDER)
        print(f"Downloaded {package} to {EXCHANGE_FOLDER.as_posix()}")


def install_all():
    """Installs all packages in requirements."""
    for package in get_package_list():
        install_from_folder(package, EXCHANGE_FOLDER)
        print(f"Installed {package} from {EXCHANGE_FOLDER.as_posix()}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--download", help="Download packages", action="store_true")
    parser.add_argument("--install", help="Install packages", action="store_true")
    args = parser.parse_args()
    if args.download:
        download_all()
    elif args.install:
        install_all()
    else:
        print("Nothing to do! either download or install!")



from typing import Optional

from flowchem.components.stdlib import ActiveComponent


class Dummy(ActiveComponent):
    """
    A fake component, used internally for testing.

    ::: warning
    This component *can* be used in a non-dry run protocol execution. It does not do anything.
    :::

    Arguments:
    - `name`: The component's name.

    Attributes:
    - `name`: The component's name.
    - `active`: Whether the component is active. This doesn't actually mean anything.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self.active = False

        self._base_state = dict(active=False)

    async def _update(self) -> None:
        if self.active:
            print("Active!")
        else:
            print("Inactive.")



from ord_schema.proto.reaction_pb2 import ReactionInput, FlowRate
# from ord_schema.proto.reaction_pb2 import Compound, CrudeComponent, ReactionRole, ReactionIdentifier, CompoundPreparation,
from flowchem.units import flowchem_ureg


def add_flowrate_to_input(reaction: ReactionInput, flowrate_text: str):
    """ Add a flowrate to a reaction. """

    # Parse the flowrate
    flowrate = flowchem_ureg(flowrate_text)
    assert flowrate.units == flowchem_ureg.volume / flowchem_ureg.time, "Flowrate must be in units of volume/time"

    # Convert it to ml/min (we could use different ORD values, but this is easier)
    flowrate.ito(flowchem_ureg.milliliter / flowchem_ureg.minute)

    # Make it into a protobuf
    flow_rate_pb = FlowRate()
    flow_rate_pb.value = flowrate.magnitude
    flow_rate_pb.units = FlowRate.FlowRateUnit.Value("MILLILITER_PER_MINUTE")

    # Add the flowrate to the reaction
    reaction.flow_rate.CopyFrom(flow_rate_pb)

    return reaction



"""
Driver for Huber chillers.
"""
import asyncio
import warnings
from loguru import logger
from dataclasses import dataclass
from typing import List, Dict, Optional

import aioserial
import pint

from flowchem.components.stdlib import TempControl
from flowchem.exceptions import InvalidConfiguration, DeviceError
from flowchem.units import flowchem_ureg


@dataclass
class PBCommand:
    """Class representing a PBCommand"""

    command: str

    def to_chiller(self) -> bytes:
        """Validate and encode to bytes array to be transmitted."""
        self.validate()
        return self.command.encode("ascii")

    def validate(self):
        """Check command structure to be compliant with PB format"""
        if len(self.command) == 8:
            self.command += "\r\n"
        # 10 characters
        assert len(self.command) == 10
        # Starts with {
        assert self.command[0] == "{"
        # M for master (commands) S for slave (replies).
        assert self.command[1] in ("M", "S")
        # Address, i.e. the desired function. Hex encoded.
        assert 0 <= int(self.command[2:4], 16) < 256
        # Value
        assert self.command[4:8] == "****" or 0 <= int(self.command[4:8], 16) <= 65536
        # EOL
        assert self.command[8:10] == "\r\n"

    @property
    def data(self) -> str:
        """Data portion of PBCommand."""
        return self.command[4:8]

    def parse_temperature(self) -> str:
        """Parse a device temp from hex string to celsius float [two's complement 16 bit signed hex, see manual]"""
        temp = (
            (int(self.data, 16) - 65536) / 100
            if int(self.data, 16) > 32767
            else (int(self.data, 16)) / 100
        )
        # -151 used for invalid temperatures
        if temp == -151:
            return ""
        return str(flowchem_ureg(f"{temp} °C"))

    def parse_integer(self) -> int:
        """Parse a device reply from hexadecimal string to base 10 integers."""
        return int(self.data, 16)

    def parse_rpm(self) -> str:
        """ Parse a device reply from hexadecimal string to rpm. """
        return str(flowchem_ureg(f"{self.parse_integer()} rpm"))

    def parse_bits(self) -> List[bool]:
        """ " Parse a device reply from hexadecimal string to 16 constituting bits."""
        bits = f"{int(self.data, 16):016b}"
        return [bool(int(x)) for x in bits]

    def parse_boolean(self):
        """ " Parse a device reply from hexadecimal string (0x0000 or 0x0001) to boolean."""
        return self.parse_integer() == 1

    def parse_status1(self) -> Dict[str, bool]:
        """Parse response to status1 command and returns dict"""
        bits = self.parse_bits()
        return dict(
            temp_ctl_is_process=bits[0],
            circulation_active=bits[1],
            refrigerator_on=bits[2],
            temp_is_process=bits[3],
            circulating_pump=bits[4],
            cooling_power_available=bits[5],
            tkeylock=bits[6],
            is_pid_auto=bits[7],
            error=bits[8],
            warning=bits[9],
            int_temp_mode=bits[10],
            ext_temp_mode=bits[11],
            dv_e_grade=bits[12],
            power_failure=bits[13],
            freeze_protection=bits[14],
        )

    def parse_status2(self) -> Dict[str, bool]:
        """Parse response to status2 command and returns dict. See manufacturer docs for more info"""
        bits = self.parse_bits()
        return dict(
            controller_is_external=bits[0],
            drip_tray_full=bits[5],
            venting_active=bits[7],
            venting_successful=bits[8],
            venting_monitored=bits[9],
        )


class HuberChiller(TempControl):
    """
    Control class for Huber chillers.
    """

    def __init__(self, aio: aioserial.AioSerial, name=None):
        super().__init__(name)
        self._serial = aio

    @classmethod
    def from_config(cls, port, name=None, **serial_kwargs):
        """
        Create instance from config dict. Used by server to initialize obj from config.

        Only required parameter is 'port'. Optional 'loop' + others (see AioSerial())
        """
        try:
            serial_object = aioserial.AioSerial(port, **serial_kwargs)
        except aioserial.SerialException as e:
            raise InvalidConfiguration(
                f"Cannot connect to the HuberChiller on the port <{port}>"
            ) from e

        return cls(serial_object, name)

    async def initialize(self):
        """Ensure the connection w/ device is working."""
        sn = await self.serial_number()
        if sn == 0:
            raise DeviceError("No reply received from Huber Chiller!")
        logger.debug(f"Connected with Huber Chiller S/N {sn}")

    async def send_command_and_read_reply(self, command: str) -> str:
        """Sends a command to the chiller and reads the reply.

        :param command: string to be transmitted
        :return: reply received
        """
        # Send command. Using PBCommand ensure command validation, see PBCommand.to_chiller()
        pb_command = PBCommand(command.upper())
        await self._serial.write_async(pb_command.to_chiller())
        logger.debug(f"Command {command[0:8]} sent to chiller!")

        # Receive reply and return it after decoding
        try:
            reply = await asyncio.wait_for(self._serial.readline_async(), 1)
        except asyncio.TimeoutError:
            warnings.warn(
                "No reply received. Likely the command is not supported by the hardware!"
            )
            logger.error("No reply received")
            return command.replace("M", "S").replace(
                "****", "0000"
            )  # Fake reply to keep going

        logger.debug(f"Reply {reply[0:8].decode('ascii')} received")
        return reply.decode("ascii")

    async def get_temperature_setpoint(self) -> str:
        """Returns the set point used by temperature controller. Internal if not probe, otherwise process temp."""
        reply = await self.send_command_and_read_reply("{M00****")
        return PBCommand(reply).parse_temperature()

    async def set_temperature_setpoint(self, temp: str):
        """ Set the set point used by temperature controller. Internal if not probe, otherwise process temp. """
        min_t = flowchem_ureg(await self.min_setpoint())
        max_t = flowchem_ureg(await self.max_setpoint())
        temp = flowchem_ureg(temp)

        if temp > max_t:
            temp = max_t
            warnings.warn(
                f"Temperature requested {temp} is out of range [{min_t} - {max_t}] for HuberChiller {self}!"
                f"Setting to {max_t} instead."
            )
        if temp < min_t:
            temp = min_t
            warnings.warn(
                f"Temperature requested {temp} is out of range [{min_t} - {max_t}] for HuberChiller {self}!"
                f"Setting to {min_t} instead."
            )

        await self.send_command_and_read_reply("{M00" + self._temp_to_string(temp))

    async def internal_temperature(self) -> str:
        """Returns internal temp (bath temperature)."""
        reply = await self.send_command_and_read_reply("{M01****")
        return PBCommand(reply).parse_temperature()

    async def process_temperature(self) -> str:
        """Returns the current process temperature. If not T probe, the device returns -151, here parsed as None."""
        reply = await self.send_command_and_read_reply("{M3A****")
        return PBCommand(reply).parse_temperature()

    async def return_temperature(self) -> str:
        """Returns the temp of the thermal fluid flowing back to the device."""
        reply = await self.send_command_and_read_reply("{M02****")
        return PBCommand(reply).parse_temperature()

    async def pump_pressure(self) -> str:
        """Return pump pressure in mbar (note that you probably want barg, i.e. to remove 1 bar)"""
        reply = await self.send_command_and_read_reply("{M03****")
        pressure = PBCommand(reply).parse_integer()
        return str(flowchem_ureg(f"{pressure} mbar"))

    async def current_power(self) -> str:
        """Returns the current power in Watts (negative for cooling, positive for heating)."""
        reply = await self.send_command_and_read_reply("{M04****")
        power = PBCommand(reply).parse_integer()
        return str(flowchem_ureg(f"{power} watt"))

    async def status(self) -> Dict[str, bool]:
        """Returns the info contained in vstatus1 as dict."""
        reply = await self.send_command_and_read_reply("{M0A****")
        return PBCommand(reply).parse_status1()

    async def status2(self) -> Dict[str, bool]:
        """Returns the info contained in vstatus2 as dict."""
        reply = await self.send_command_and_read_reply("{M3C****")
        return PBCommand(reply).parse_status2()

    async def is_temperature_control_active(self) -> bool:
        """Returns whether temperature control is active or not."""
        reply = await self.send_command_and_read_reply("{M14****")
        return PBCommand(reply).parse_boolean()

    async def start_temperature_control(self):
        """Starts temperature control, i.e. start operation."""
        await self.send_command_and_read_reply("{M140001")

    async def stop_temperature_control(self):
        """Stops temperature control, i.e. stop operation."""
        await self.send_command_and_read_reply("{M140000")

    async def is_circulation_active(self) -> bool:
        """Returns whether temperature control is active or not."""
        reply = await self.send_command_and_read_reply("{M16****")
        return PBCommand(reply).parse_boolean()

    async def start_circulation(self):
        """Starts circulation pump."""
        await self.send_command_and_read_reply("{M160001")

    async def stop_circulation(self):
        """Stops circulation pump."""
        await self.send_command_and_read_reply("{M160000")

    async def pump_speed(self) -> str:
        """Returns current circulation pump speed (in rpm)."""
        reply = await self.send_command_and_read_reply("{M26****")
        return PBCommand(reply).parse_rpm()

    async def pump_speed_setpoint(self) -> str:
        """Returns the set point of the circulation pump speed (in rpm)."""
        reply = await self.send_command_and_read_reply("{M48****")
        return PBCommand(reply).parse_rpm()

    async def set_pump_speed(self, rpm: str):
        """Set the pump speed, in rpm. See device display for range."""
        parsed_rpm = flowchem_ureg(rpm)
        await self.send_command_and_read_reply("{M48" + self._int_to_string(parsed_rpm.m_as("rpm")))

    async def cooling_water_temp(self) -> str:
        """Returns the cooling water inlet temperature (in Celsius)."""
        reply = await self.send_command_and_read_reply("{M2C****")
        return PBCommand(reply).parse_temperature()

    async def cooling_water_pressure(self) -> Optional[float]:
        """Returns the cooling water inlet pressure (in mbar)."""
        reply = await self.send_command_and_read_reply("{M2D****")
        if pressure := PBCommand(reply).parse_integer() == 64536:
            return None
        else:
            return pressure

    async def cooling_water_temp_outflow(self) -> str:
        """Returns the cooling water outlet temperature (in Celsius)."""
        reply = await self.send_command_and_read_reply("{M4C****")
        return PBCommand(reply).parse_temperature()

    async def min_setpoint(self) -> str:
        """Returns the minimum accepted value for the temperature setpoint (in Celsius)."""
        reply = await self.send_command_and_read_reply("{M30****")
        return PBCommand(reply).parse_temperature()

    async def max_setpoint(self) -> str:
        """Returns the maximum accepted value for the temperature setpoint (in Celsius)."""
        reply = await self.send_command_and_read_reply("{M31****")
        return PBCommand(reply).parse_temperature()

    async def alarm_max_internal_temp(self) -> str:
        """Returns the max internal temp before the alarm is triggered and a fault generated."""
        reply = await self.send_command_and_read_reply("{M51****")
        return PBCommand(reply).parse_temperature()

    async def set_alarm_max_internal_temp(self, temp: str):
        """ Sets the max internal temp before the alarm is triggered and a fault generated. """
        temp = flowchem_ureg(temp)
        await self.send_command_and_read_reply("{M51" + self._temp_to_string(temp))

    async def alarm_min_internal_temp(self) -> str:
        """Returns the min internal temp before the alarm is triggered and a fault generated."""
        reply = await self.send_command_and_read_reply("{M52****")
        return PBCommand(reply).parse_temperature()

    async def set_alarm_min_internal_temp(self, temp: str):
        """ Sets the min internal temp before the alarm is triggered and a fault generated. """
        temp = flowchem_ureg(temp)
        await self.send_command_and_read_reply("{M52" + self._temp_to_string(temp))

    async def alarm_max_process_temp(self) -> str:
        """Returns the max process temp before the alarm is triggered and a fault generated."""
        reply = await self.send_command_and_read_reply("{M53****")
        return PBCommand(reply).parse_temperature()

    async def set_alarm_max_process_temp(self, temp: str):
        """ Sets the max process temp before the alarm is triggered and a fault generated. """
        temp = flowchem_ureg(temp)
        await self.send_command_and_read_reply("{M53" + self._temp_to_string(temp))

    async def alarm_min_process_temp(self) -> str:
        """Returns the min process temp before the alarm is triggered and a fault generated."""
        reply = await self.send_command_and_read_reply("{M54****")
        return PBCommand(reply).parse_temperature()

    async def set_alarm_min_process_temp(self, temp: str):
        """ Sets the min process temp before the alarm is triggered and a fault generated. """
        temp = flowchem_ureg(temp)
        await self.send_command_and_read_reply("{M54" + self._temp_to_string(temp))

    async def set_ramp_duration(self, ramp_time: str):
        """ Sets the duration (in seconds) of a ramp to the temperature set by a later call to ramp_to_temperature. """
        parsed_time = flowchem_ureg(ramp_time)
        await self.send_command_and_read_reply("{M59" + self._int_to_string(parsed_time.m_as("s")))

    async def ramp_to_temperature(self, temperature: str):
        """ Sets the duration (in seconds) of a ramp to the temperature set by a later call to start_ramp(). """
        temp = flowchem_ureg(temperature)
        await self.send_command_and_read_reply(
            "{M5A" + self._temp_to_string(temp)
        )

    async def is_venting(self) -> bool:
        """Whether the chiller is venting or not."""
        reply = await self.send_command_and_read_reply("{M6F****")
        return PBCommand(reply).parse_boolean()

    async def start_venting(self):
        """Starts venting. ONLY USE DURING SETUP! READ THE MANUAL!"""
        await self.send_command_and_read_reply("{M6F0001")

    async def stop_venting(self):
        """Stops venting."""
        await self.send_command_and_read_reply("{M6F0000")

    async def is_draining(self) -> bool:
        """Whether the chiller is venting or not."""
        reply = await self.send_command_and_read_reply("{M70****")
        return PBCommand(reply).parse_boolean()

    async def start_draining(self):
        """Starts venting. ONLY USE DURING SHUT DOWN! READ THE MANUAL!"""
        await self.send_command_and_read_reply("{M700001")

    async def stop_draining(self):
        """Stops venting."""
        await self.send_command_and_read_reply("{M700000")

    async def serial_number(self) -> int:
        """GGet serial number."""
        s1 = await self.send_command_and_read_reply("{M1B****")
        s2 = await self.send_command_and_read_reply("{M1C****")
        pb1, pb2 = PBCommand(s1), PBCommand(s2)
        return int(pb1.data + pb2.data, 16)

    async def wait_for_temperature_simple(self) -> None:
        """Returns as soon as the target temperature range has been reached, or timeout."""
        raise NotImplementedError

    async def wait_for_temperature_stable(self) -> None:
        """Returns when the target temperature range has been maintained for X seconds, or timeout."""
        raise NotImplementedError

    @staticmethod
    def _temp_to_string(temp: pint.Quantity) -> str:
        """ From temperature to string for command. f^-1 of PCommand.parse_temperature. """
        minT = flowchem_ureg.Quantity(-151, "celsius")
        maxT = flowchem_ureg.Quantity(327, "celsius")
        assert minT <= temp <= maxT
        # Hexadecimal two's complement
        return f"{int(temp.m_as('°C') * 100) & 65535:04X}"

    @staticmethod
    def _int_to_string(number: int) -> str:
        """ From temperature to string for command. f^-1 of PCommand.parse_integer. """
        return f"{number:04X}"

    async def __aenter__(self):
        await self.initialize()
        await self.set_temperature_setpoint(temp="20 degC")
        await self.start_temperature_control()
        await self.start_circulation()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.set_temperature_setpoint("20 C")

        # Wait until close to room temperature before turning off chiller
        while flowchem_ureg.parse_expression(await self.process_temperature()) > flowchem_ureg.parse_expression("40 degC"):
            await asyncio.sleep(5)

        # Actually turn off chiller
        await self.stop_circulation()
        await self.stop_temperature_control()

    async def _update(self):
        await self.set_temperature_setpoint(self.temp)

    def get_router(self):
        """Creates an APIRouter for this HuberChiller instance."""
        # Local import to allow direct use of HuberChiller w/o fastapi installed
        from fastapi import APIRouter

        router = APIRouter()
        router.add_api_route(
            "/temperature/set-point", self.get_temperature_setpoint, methods=["GET"]
        )
        router.add_api_route(
            "/temperature/set-point", self.set_temperature_setpoint, methods=["PUT"]
        )
        router.add_api_route(
            "/temperature/set-point/min", self.min_setpoint, methods=["GET"]
        )
        router.add_api_route(
            "/temperature/set-point/max", self.max_setpoint, methods=["GET"]
        )
        router.add_api_route(
            "/temperature/process", self.process_temperature, methods=["GET"]
        )
        router.add_api_route(
            "/temperature/internal", self.internal_temperature, methods=["GET"]
        )
        router.add_api_route(
            "/temperature/return", self.return_temperature, methods=["GET"]
        )
        router.add_api_route("/power-exchanged", self.current_power, methods=["GET"])
        router.add_api_route("/status", self.status, methods=["GET"])
        router.add_api_route("/status2", self.status2, methods=["GET"])
        router.add_api_route("/pump/speed", self.pump_speed, methods=["GET"])
        router.add_api_route(
            "/temperature-control", self.is_temperature_control_active, methods=["GET"]
        )
        router.add_api_route(
            "/temperature-control/start",
            self.start_temperature_control,
            methods=["GET"],
        )
        router.add_api_route(
            "/temperature-control/stop", self.stop_temperature_control, methods=["GET"]
        )
        router.add_api_route(
            "/pump/circulation", self.is_circulation_active, methods=["GET"]
        )
        router.add_api_route(
            "/pump/circulation/start", self.start_circulation, methods=["GET"]
        )
        router.add_api_route(
            "/pump/circulation/stop", self.stop_circulation, methods=["GET"]
        )
        router.add_api_route("/pump/pressure", self.pump_pressure, methods=["GET"])
        router.add_api_route("/pump/speed", self.pump_speed, methods=["GET"])
        router.add_api_route(
            "/pump/speed/set-point", self.pump_speed_setpoint, methods=["GET"]
        )
        router.add_api_route(
            "/pump/speed/set-point", self.set_pump_speed, methods=["PUT"]
        )
        router.add_api_route(
            "/cooling-water/temperature-inlet", self.cooling_water_temp, methods=["GET"]
        )
        router.add_api_route(
            "/cooling-water/temperature-outlet",
            self.cooling_water_temp_outflow,
            methods=["GET"],
        )
        router.add_api_route(
            "/cooling-water/pressure", self.cooling_water_pressure, methods=["GET"]
        )
        router.add_api_route(
            "/alarm/process/min-temp", self.alarm_min_process_temp, methods=["GET"]
        )
        router.add_api_route(
            "/alarm/process/max-temp", self.alarm_max_process_temp, methods=["GET"]
        )
        router.add_api_route(
            "/alarm/process/min-temp", self.set_alarm_min_process_temp, methods=["PUT"]
        )
        router.add_api_route(
            "/alarm/process/max-temp", self.set_alarm_min_process_temp, methods=["PUT"]
        )
        router.add_api_route(
            "/alarm/internal/min-temp", self.alarm_min_internal_temp, methods=["GET"]
        )
        router.add_api_route(
            "/alarm/internal/max-temp", self.alarm_max_internal_temp, methods=["GET"]
        )
        router.add_api_route(
            "/alarm/internal/min-temp",
            self.set_alarm_min_internal_temp,
            methods=["PUT"],
        )
        router.add_api_route(
            "/alarm/internal/max-temp",
            self.set_alarm_min_internal_temp,
            methods=["PUT"],
        )
        router.add_api_route("/venting/is_venting", self.is_venting, methods=["GET"])
        router.add_api_route("/venting/start", self.start_venting, methods=["GET"])
        router.add_api_route("/venting/stop", self.stop_venting, methods=["GET"])
        router.add_api_route("/draining/is_venting", self.is_draining, methods=["GET"])
        router.add_api_route("/draining/start", self.start_draining, methods=["GET"])
        router.add_api_route("/draining/stop", self.stop_draining, methods=["GET"])
        router.add_api_route("/serial_number", self.serial_number, methods=["GET"])

        return router


if __name__ == "__main__":
    chiller = HuberChiller(aioserial.AioSerial(port="COM8"))
    status = asyncio.run(chiller.status())
    print(status)



""" Unit-conversion related functions """
import os

import pint

flowchem_ureg = pint.UnitRegistry(autoconvert_offset_to_baseunit=True)

UNIT_REGISTER = os.path.join(os.path.dirname(os.path.realpath(__file__)), "units.txt")
flowchem_ureg.load_definitions(UNIT_REGISTER)



from .AzuraCompactPump import AzuraCompactPump
from .KnauerValve import (
    Knauer12PortValve,
    Knauer16PortValve,
    Knauer6Port6PositionValve,
    Knauer6Port2PositionValve,
)
from .Knauer_autodiscover import autodiscover_knauer



from __future__ import annotations

import asyncio
import json
import os
import time
from hashlib import blake2b
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from warnings import warn
from loguru import logger

import aiofiles
import ipywidgets as widgets
from bokeh.io import output_notebook, push_notebook, show
from bokeh.plotting import figure
from bokeh.resources import INLINE
from IPython import get_ipython
from IPython.display import display

from flowchem.components.stdlib import ActiveComponent, Sensor
from flowchem.core.execute import main

if TYPE_CHECKING:
    from flowchem import Protocol, Datapoint


class Experiment(object):
    """
    Experiments contain all data from execution of a protocol.

    Arguments:
    - `protocol`: The protocol for which the experiment was conducted.
    - `compiled_protocol`: The results of `protocol._compile()`.
    - `verbosity`: See `Protocol.execute` for a description of the verbosity options.
    - `dry_run`: Whether the experiment is a dry run and, if so, by what factor it is sped up by.

    Attributes:
    - `apparatus`: The apparatus upon which the experiment is conducted.
    - `cancelled`: Whether the experiment is cancelled.
    - `compiled_protocol`: The results of `protocol._compile()`.
    - `data`: A list of `Datapoint` namedtuples from the experiment's sensors.
    - `dry_run`: Whether the experiment is a dry run and, if so, by what factor it is sped up by.
    - `end_time`: The Unix time of the experiment's end.
    - `executed_procedures`: A list of the procedures that were executed during the experiment.
    - `experiment_id`: The experiment's ID. By default, of the form `YYYY_MM_DD_HH_MM_SS_HASH`, where HASH is the 3-bytes hexadecimal blake2b hash of the protocol's YAML.
    - `paused`: Whether the experiment is currently paused.
    - `protocol`: The protocol for which the experiment was conducted.
    - `start_time`: The Unix time of the experiment's is.
    """

    def __init__(self, protocol: "Protocol"):
        # args
        self.apparatus = protocol.apparatus
        self.protocol = protocol

        # computed values
        self.experiment_id: Optional[str] = None

        # default values
        self.dry_run: Union[bool, int]
        self.start_time: float  # hasn't started until main() is called
        self.created_time = time.time()  # when the object was created (might be diff)
        self.end_time: float
        self.data: Dict[str, List[Datapoint]] = {}
        self.cancelled = False
        self.was_executed = False
        self.executed_procedures: List[
            Dict[str, Union[float, Dict[str, Any], str, ActiveComponent]]
        ] = []

        # internal values (unstable!)
        _local_time = time.localtime(self.created_time)
        self._created_time_local: str = time.strftime("%Y_%m_%d_%H_%M_%S", _local_time)
        self._charts = {}  # type: ignore
        self._graphs_shown = False
        self._sensors = self.apparatus[Sensor]
        self._sensors.reverse()
        self._device_name_to_unit = {c.name: c._unit for c in self._sensors}
        self._sensor_names: List[str] = [s.name for s in self._sensors]
        self._bound_logger = None
        self._plot_height = 300
        self._is_executing = False
        self._paused = False
        self._pause_times: List[Dict[str, float]] = []
        self._end_loop = False  # when to stop monitoring the buttons
        self._file_logger_id: Optional[int] = None
        self._log_file: Optional[Path] = None
        self._data_file: Optional[Path] = None
        self._transformed_data: Dict[str, Dict[str, List[Datapoint]]] = {
            s: {"datapoints": [], "timestamps": []} for s in self._sensor_names
        }

    def __str__(self):
        return f"Experiment {self.experiment_id}"

    def __repr__(self):
        return f"<Experiment {self.experiment_id}>"

    async def _update(self, device: str, datapoint):

        # If a chart has been registered to the device, update it.
        if device not in self.data:
            self.data[device] = []
        self.data[device].append(datapoint)

        if self._data_file is not None:
            line = json.dumps(
                {
                    "device": device,
                    "timestamp": datapoint.timestamp,
                    "experiment_elapsed_time": datapoint.experiment_elapsed_time,
                    "data": datapoint.data,
                    "unit": self.protocol.apparatus[device]._unit,
                }
            )
            async with aiofiles.open(self._data_file, "a+") as f:
                await f.write(line + "\n")

        if get_ipython() is None:
            return

        if not self._graphs_shown:
            logger.debug("Graphs not shown. Initializing...")
            for sensor, output in self._sensor_outputs.items():  # type: ignore
                logger.trace(f"Initializing graph for {sensor}")

                # bind the height of the graph to the selected plot height
                output.layout.height = f"{self._plot_height}px"

                with output:
                    # create the figure object
                    p = figure(
                        title=f"{sensor} data",
                        plot_height=self._plot_height,
                        plot_width=600,
                    )
                    r = p.line(
                        source=self._transformed_data[sensor.name],
                        x="timestamps",
                        y="datapoints",
                        color="#2222aa",
                        line_width=3,
                    )
                    p.xaxis.axis_label = "Experiment elapsed time (seconds)"
                    p.yaxis.axis_label = self._device_name_to_unit[sensor.name]

                    # since we're in the with-statement, this will show up in the accordion
                    output_notebook(resources=INLINE, hide_banner=True)
                    target = show(p, notebook_handle=True)

                    # save the target and plot for later updating
                    self._charts[sensor.name] = (target, r)
                logger.trace(f"Sucessfully initialized graph for {sensor.name}")
            logger.trace("All graphs successfully initialized")
            self._graphs_shown = True

        if device in self._transformed_data:
            target, r = self._charts[device]
            self._transformed_data[device]["datapoints"].append(datapoint.data)
            self._transformed_data[device]["timestamps"].append(
                datapoint.experiment_elapsed_time
            )
            r.data_source.data["datapoints"] = self._transformed_data[device][
                "datapoints"
            ]
            r.data_source.data["timestamps"] = self._transformed_data[device][
                "timestamps"
            ]
            push_notebook(handle=target)

    def _on_stop_clicked(self, b):
        logger.debug("Stop button pressed.")
        self.cancelled = True

    def _on_pause_clicked(self, b):
        self.paused = not self.paused

    @property
    def _total_paused_duration(self) -> float:
        """Calculate the total amount of time the experiment was paused for."""
        duration = 0.0
        for pause in self._pause_times:
            if "stop" in pause:
                duration += pause["stop"] - pause["start"]
        return duration

    def get_confirmation(self):
        """ Ensure user input is present before starting procedure. """
        confirmation = input("Execute? [y/N]: ").lower()
        if not confirmation or confirmation[0] != "y":
            logger.critical("Aborting execution...")
            raise RuntimeError("Execution aborted by user.")

    def _execute(
        self,
        dry_run: Union[bool, int],
        verbosity: str,
        confirm: bool,
        strict: bool,
        log_file: Union[str, bool, os.PathLike, None],
        log_file_verbosity: Optional[str],
        log_file_compression: Optional[str],
        data_file: Union[str, bool, os.PathLike, None],
    ):
        self.dry_run = dry_run

        # make the user confirm if it's the real deal
        if not self.dry_run and not confirm:
            self.get_confirmation()

        self._compiled_protocol = self.protocol._compile(dry_run=bool(dry_run))

        # now that we're ready to start, create the time and ID attributes
        # protocol_hash: str = blake2b(str(self.protocol.yaml()).encode(), digest_size=3).hexdigest()
        protocol_hash = "329472389fhsd"
        self.experiment_id = f"{self._created_time_local}_{protocol_hash}"

        # handle logging to a file
        if log_file:
            # automatically log to the mw directory
            if log_file is True:
                mw_path = Path("~/.flowchem").expanduser()
                try:
                    mw_path.mkdir()
                except FileExistsError:
                    pass
                log_file = mw_path / Path(self.experiment_id + ".log.jsonl")

            # for typing's sake
            assert isinstance(log_file, (str, os.PathLike))

            # automatically configure a logger to persist the logs
            self._file_logger_id = logger.add(
                log_file,
                level=verbosity.upper()
                if log_file_verbosity is None
                else log_file_verbosity.upper(),
                compression=log_file_compression,
                serialize=True,
                enqueue=True,
            )
            logger.trace(f"File logger ID is {self._file_logger_id}")

            # for typing's sake
            assert isinstance(log_file, (str, os.PathLike))

            # determine the log file's path
            if log_file_compression is not None:
                self._log_file = Path(log_file)
                self._log_file = self._log_file.with_suffix(
                    self._log_file.suffix + "." + log_file_compression
                )
            else:
                self._log_file = Path(log_file)

        if data_file:
            # automatically log to user directory
            if data_file is True:
                mw_path = Path("~/.flowchem").expanduser()
                try:
                    mw_path.mkdir()
                except FileExistsError:
                    pass
                self._data_file = mw_path / Path(self.experiment_id + ".data.jsonl")
            elif isinstance(data_file, (str, os.PathLike)):
                self._data_file = Path(data_file)
            else:
                raise TypeError(
                    f"Invalid type {type(data_file)} for data file."
                    "Expected str or os.PathLike (such as a pathlib.Path object)."
                )

        if get_ipython():
            self._display(verbosity=verbosity.upper(), strict=strict)
            asyncio.ensure_future(main(experiment=self, dry_run=dry_run, strict=strict))
        else:
            asyncio.run(main(experiment=self, dry_run=dry_run, strict=strict))

    def _display(self, verbosity: str, strict: bool):

        # create pause button
        self._pause_button = widgets.Button(description="Pause", icon="pause")
        self._pause_button.on_click(self._on_pause_clicked)

        # create a stop button
        self._stop_button = widgets.Button(
            description="Stop", button_style="danger", icon="stop"
        )
        self._stop_button.on_click(self._on_stop_clicked)

        # create a nice, pretty HTML string wth the metadata
        metadata = "<ul>"
        for k, v in {
            "Apparatus": self.apparatus.name,
            "Protocol": self.protocol.name,
            "Description": self.protocol.description,
            "Start time": time.ctime(self.created_time),
            "Expected completion": time.ctime(
                self.created_time + self.protocol._inferred_duration
            ),
            "Procedure count": sum([len(x) for x in self._compiled_protocol.values()]),
            "Abort on error": strict,
            "Log file": self._log_file.absolute() if self._log_file else None,
            "Data file": self._data_file.absolute() if self._data_file else None,
        }.items():
            if not v:
                continue
            metadata += f"<li><b>{k}:</b> {v}</li>"
        metadata += "</ul>"

        # create the output tab widget with a log tab
        self._tab = widgets.Tab()
        self._log_widget = widgets.Output()
        self._tab.children = (self._log_widget, widgets.HTML(value=metadata))
        self._tab.set_title(0, "Log")
        self._tab.set_title(1, "Metadata")

        if self._sensors:
            self._sensor_outputs = {s: widgets.Output() for s in self._sensors}

            self._accordion = widgets.Accordion(
                children=tuple(self._sensor_outputs.values())
            )
            self._tab.children = tuple(list(self._tab.children) + [self._accordion])
            self._tab.set_title(2, "Sensors")

            # we know that the accordion will line up with the dict since dict order
            # is preserved in Python 3.7+
            for i, sensor in enumerate(self._sensor_outputs):
                self._accordion.set_title(i, sensor.name)

        # decide whether to show a pause button
        buttons = [self._stop_button]
        if type(self.dry_run) != int:
            buttons.insert(0, self._pause_button)

        self._output_widget = widgets.VBox(
            [
                widgets.HTML(value=f"<h3>Experiment {self.experiment_id}</h3>"),
                widgets.HBox(buttons),
                self._tab,
            ]
        )

        def _log(x):
            with self._log_widget:  # the log
                # .xxx in floats
                pad_length = len(str(int(self.protocol._inferred_duration))) + 4
                # we don't (cleanup) to look weird, so pad to at least its length
                pad_length = max((pad_length, len("cleanup")))

                if self.is_executing and not self.was_executed:
                    elapsed_time = f"{time.time() - self.start_time:0{pad_length}.3f}"
                    print(f"({elapsed_time}) {x.rstrip()}")
                elif self.was_executed:
                    print(f"({'cleanup'.center(pad_length)}) {x.rstrip()}")
                else:
                    print(f"({'setup'.center(pad_length)}) " + x.rstrip())

        # don't enqueue since it breaks the graphing
        self._bound_logger = logger.add(
            lambda x: _log(x),
            level=verbosity,
            colorize=True,
            format="{level.icon} {message}",
        )  # type: ignore

        display(self._output_widget)

    @property
    def is_executing(self):
        return self._is_executing

    @is_executing.setter
    def is_executing(self, is_executing):
        if not is_executing and self._file_logger_id is not None:
            logger.info("Wrote logs to " + str(self._log_file.absolute()))
            logger.trace(f"Removing generated file logger {self._file_logger_id}")
            logger.remove(self._file_logger_id)
            logger.trace("File logger removed")
            # ensure that an execution without logging after one with it doesn't break
            self._log_file = None
            self._file_logger_id = None
        if not is_executing and self._data_file:
            logger.info("Wrote data to " + str(self._data_file.absolute()))
            logger._data_file = None

        # this deactivates sensor monitoring and button usability
        if not is_executing:
            self._end_loop = True
            logger.trace(f"_end_loop for {self} is now True.")

        logger.trace(f"{repr(self)}.is_executing is now {is_executing}")
        self._is_executing = is_executing

    @property
    def paused(self):
        return self._paused

    @paused.setter
    def paused(self, paused):

        # pausing a sped up dry run is meaningless
        if type(self.dry_run) == int:
            warn("Pausing a speed run is not supported. This will have no effect.")

        # issue a warning if the user overuses the pause button
        if len(self._pause_times) >= 3:
            logger.warning("Pausing repeatedly may adversely affect protocol timing.")

        if paused and not self._paused:
            logger.warning("Paused execution.")
            self._pause_times.append(dict(start=time.time()))
        elif not paused and self._paused:
            self._pause_times[-1]["stop"] = time.time()
            logger.warning("Resumed execution.")
        self._paused = paused

        # control the pause button
        self._pause_button.description = "Resume" if paused else "Pause"
        self._pause_button.button_style = "success" if paused else ""
        self._pause_button.icon = "play" if paused else "pause"



from typing import Optional

from flowchem.components.stdlib import Sensor


class BrokenDummySensor(Sensor):
    """
    A dummy sensor returning the number of times it has been read. After a few reads, it fails, raising a `RuntimeError`.

    ::: danger
    Using this component during real protocol execution will result in a failure.
    :::

    Attributes:
    - `name` (str, optional): The name of the Sensor.
    - `rate` (Quantity): Data collection rate in Hz. A rate of 0 Hz corresponds to the sensor being off.
    """

    def __init__(self, name: Optional[str] = None):
        super().__init__(name=name)
        self._unit = "Dimensionless"
        self.counter = 0

    async def _read(self) -> int:
        """Collect the data."""
        self.counter += 1
        if self.counter > 15 and self.rate:
            raise RuntimeError("Unsurprisingly, the broken sensor is broken.")
        return self.counter
